"""
Module unit tests.
"""

from l5x import module
from tests import fixture
import unittest


class Modules(unittest.TestCase):
    """Tests for the project's top-level modules container."""
    def setUp(self):
        self.prj = fixture.string_to_project("""<?xml version="1.0" encoding="UTF-8" standalone="yes"?>
<RSLogix5000Content SchemaRevision="1.0" SoftwareRevision="20.01" TargetName="test" TargetType="Controller" ContainsContext="false" Owner="admin" ExportDate="Mon Jul 20 02:35:01 2020" ExportOptions="DecoratedData ForceProtectedEncoding AllProjDocTrans">
<Controller Use="Target" Name="test" ProcessorType="1756-L61" MajorRev="20" MinorRev="11" TimeSlice="20" ShareUnusedTimeSlice="1" ProjectCreationDate="Sat Jul 18 23:53:16 2020" LastModifiedDate="Sat Jul 18 23:53:18 2020" SFCExecutionControl="CurrentActive" SFCRestartPosition="MostRecent"
 SFCLastScan="DontScan" ProjectSN="16#0000_0000" MatchProjectToController="false" CanUseRPIFromProducer="false" InhibitAutomaticFirmwareUpdate="0">
<Modules>
<Module Name="Local" CatalogNumber="1756-L61" Vendor="1" ProductType="14" ProductCode="54" Major="20" Minor="11" ParentModule="Local" ParentModPortId="1" Inhibited="false" MajorFault="true"
>
<EKey State="ExactMatch"/>
<Ports>
<Port Id="1" Address="0" Type="ICP" Upstream="false">
<Bus Size="10"/>
</Port>
</Ports>
</Module>
<Module Name="mod1" CatalogNumber="1756-ENBT/A" Vendor="1" ProductType="12" ProductCode="58" Major="5" Minor="1" ParentModule="Local" ParentModPortId="1" Inhibited="false" MajorFault="false"
>
<EKey State="CompatibleModule"/>
<Ports>
<Port Id="1" Address="1" Type="ICP" Upstream="true"/>
<Port Id="2" Type="Ethernet" Upstream="false">
<Bus/>
</Port>
</Ports>
<Communications CommMethod="536870914">
<Connections/>
</Communications>
<ExtendedProperties>
<public><ConfigID>4325481</ConfigID></public></ExtendedProperties>
</Module>
<Module Name="mod2" CatalogNumber="1756-ENBT/A" Vendor="1" ProductType="12" ProductCode="58" Major="5" Minor="1" ParentModule="Local" ParentModPortId="1" Inhibited="false" MajorFault="false"
>
<EKey State="CompatibleModule"/>
<Ports>
<Port Id="1" Address="2" Type="ICP" Upstream="true"/>
<Port Id="2" Type="Ethernet" Upstream="false">
<Bus/>
</Port>
</Ports>
<Communications CommMethod="536870914">
<Connections/>
</Communications>
<ExtendedProperties>
<public><ConfigID>4325481</ConfigID></public></ExtendedProperties>
</Module>
</Modules>
<Tags/>
<Programs/>
</Controller>
</RSLogix5000Content>
""")

    def test_names_read(self):
        """Ensure names attribute returns a non-empty set of strings."""
        self.assertEqual(set(self.prj.modules.names),
                         set(('Local', 'mod1', 'mod2')))

    def test_names_readonly(self):
        """Ensure an exception is raised when attempting to write to the names attribute."""
        with self.assertRaises(AttributeError):
            self.prj.modules.names = 'foo'


class ModuleStandard(unittest.TestCase):
    """Tests for a single, non-safety module instance."""
    def setUp(self):
        element = fixture.parse_xml("""<Module Name="mod1" CatalogNumber="1756-ENBT/A" Vendor="1" ProductType="12" ProductCode="58" Major="5" Minor="1" ParentModule="Local" ParentModPortId="1" Inhibited="false" MajorFault="false"
>
<EKey State="CompatibleModule"/>
<Ports>
<Port Id="1" Address="1" Type="ICP" Upstream="true"/>
<Port Id="2" Type="Ethernet" Upstream="false">
<Bus/>
</Port>
</Ports>
<Communications CommMethod="536870914">
<Connections/>
</Communications>
<ExtendedProperties>
<public><ConfigID>4325481</ConfigID></public></ExtendedProperties>
</Module>""")
        self.module = module.Module(element)

    def test_port_names_read(self):
        """Ensure names returns a list of port IDs."""
        self.assertEqual(set(self.module.ports.names), set((1, 2)))

    def test_port_names_readonly(self):
        """Ensure an exception is raised when attempting to write to the port names attribute."""
        with self.assertRaises(AttributeError):
            self.module.ports.names = 'foo'

    def test_invalid_port_index(self):
        """Ensure invalid port indices raise an exception."""
        with self.assertRaises(KeyError):
            self.module.ports[0]

    def test_port_type(self):
        """Confirm accessing a port returns a Port instance."""
        self.assertIsInstance(self.module.ports[1], module.Port)

    def test_snn_read(self):
        """Confirm reading the SNN raises an exception."""
        with self.assertRaises(TypeError):
            self.module.snn

    def test_snn_write(self):
        """Confirm writing the SNN raises an exception."""
        with self.assertRaises(TypeError):
            self.module.snn = 'foo'


class ModuleSafetySingleSNN(unittest.TestCase):
    """Tests for safety modules with a single SNN for the entire module."""
    def setUp(self):
        self.element = fixture.parse_xml("""<Module Name="Local" CatalogNumber="1756-L61S" Vendor="1" ProductType="14" ProductCode="67" Major="20" Minor="11" ParentModule="Local" ParentModPortId="1" Inhibited="false" MajorFault="true"
 SafetyNetwork="16#0000_4544_03d1_e91a">
<EKey State="ExactMatch"/>
<Ports>
<Port Id="1" Address="0" Type="ICP" Upstream="false" Width="2">
<Bus Size="10"/>
</Port>
</Ports>
</Module>""")
        self.module = module.Module(self.element)

    def test_snn_read(self):
        """Confirm reading the SNN yields the module's SNN."""
        self.assertEqual(self.module.snn, '454403d1e91a')

    def test_snn_write(self):
        """Confirm writing the SNN correctly updates the XML attribute."""
        self.module.snn = '0'
        self.assertEqual(self.element.attrib['SafetyNetwork'],
                         '16#0000_0000_0000_0000')


class ModuleSafetyPortSNN(unittest.TestCase):
    """
    Tests for safety modules with per-port SNNs. These just confirm the
    module itself has no SNN as the SNNs must be accessed via the port
    instances, which are covered in separate unit tests.
    """
    def setUp(self):
        element = fixture.parse_xml("""<Module Name="Local" CatalogNumber="1756-L83ES" Vendor="1" ProductType="14" ProductCode="213" Major="31" Minor="11" ParentModule="Local" ParentModPortId="1" Inhibited="false" MajorFault="true"
>
<EKey State="Disabled"/>
<Ports>
<Port Id="1" Address="0" Type="ICP" Upstream="false" Width="2" SafetyNetwork="16#0000_1337_d00d_0100">
<Bus Size="4"/>
</Port>
<Port Id="2" Type="Ethernet" Upstream="false" SafetyNetwork="16#0000_1337_d00d_0101">
<Bus/>
</Port>
</Ports>
</Module>""")
        self.module = module.Module(element)

    def test_snn_read(self):
        """Confirm reading the SNN raises an exception."""
        with self.assertRaises(TypeError):
            self.module.snn

    def test_snn_write(self):
        """Confirm writing the SNN raises an exception."""
        with self.assertRaises(TypeError):
            self.module.snn = 'foo'


class Port(object):
    """Base class with common test cases for all port types."""
    def setUp(self):
        element = fixture.parse_xml(self.xml)
        self.port = module.Port(element)

    def test_type_read(self):
        """Type attribute return a the current attribute value."""
        self.assertEqual(self.port.type, self.port.element.attrib['Type'])

    def test_type_write(self):
        """Attempting to modify port type should raise an exception."""
        with self.assertRaises(AttributeError):
            self.ports.type = 'foo'

    def test_address_read(self):
        """Confirm the current address is returned via the address attribute."""
        self.assertEqual(self.port.address, self.port.element.attrib['Address'])

    def test_address_write(self):
        """Confirm correct attribute is updated when setting a new address."""
        self.port.address = '42'
        self.assertEqual(self.port.element.attrib['Address'], '42')


class PortStandard(Port, unittest.TestCase):
    """Tests for a non-safety module communication port."""
    xml = r"""<Port Id="2" Address="192.168.0.1" Type="Ethernet" Upstream="false">
<Bus/>
</Port>"""

    def test_snn_read(self):
        """Confirm reading the SNN raises an exception."""
        with self.assertRaises(TypeError):
            self.port.snn

    def test_snn_write(self):
        """Confirm writing the SNN raises an exception."""
        with self.assertRaises(TypeError):
            self.port.snn = 'foo'


class PortNoNAT(Port, unittest.TestCase):
    """Tests for ports that do not use network address translation."""
    xml = r"""<Port Id="2" Address="192.168.0.1" Type="Ethernet" Upstream="false">
<Bus/>
</Port>"""

    def test_nat_address_read(self):
        """Confirm reading the NAT address returns None."""
        self.assertIsNone(self.port.nat_address)

    def test_nat_address_write(self):
        """Confirm an exception is raised when writing the NAT address."""
        with self.assertRaises(TypeError):
            self.port.nat_address = 'foo'


class PortNAT(Port, unittest.TestCase):
    """Test for ports configured with network address translation."""
    xml = r"""<Port Id="2" Address="192.168.0.1" Type="Ethernet" Upstream="true" NATActualAddress="10.0.0.1"/>"""

    def test_nat_address_read(self):
        """Confirm reading the NAT address returns the attribute value."""
        self.assertEqual(self.port.nat_address,
                         self.port.element.attrib['NATActualAddress'])

    def test_nat_address_write(self):
        """Confirm writing the NAT address updates the XML attribute."""
        self.port.nat_address = 'foo'
        self.assertEqual(self.port.element.attrib['NATActualAddress'], 'foo')

    def test_nat_address_write_non_string(self):
        """Confirm writing a non-string raises an exception."""
        with self.assertRaises(TypeError):
            self.port.nat_address = None


class PortSafety(Port, unittest.TestCase):
    """Tests for a safety module communication port."""
    xml = r"""<Port Id="1" Address="0" Type="ICP" Upstream="false" Width="2" SafetyNetwork="16#0000_1337_d00d_0100">
<Bus Size="4"/>
</Port>"""

    def test_snn_read(self):
        """Confirm reading the SNN."""
        self.assertEqual(self.port.snn, '1337d00d0100')

    def test_snn_write(self):
        """Confirm writing the SNN correctly updates the XML attribute."""
        self.port.snn = '0'
        self.assertEqual(self.port.element.attrib['SafetyNetwork'],
                         '16#0000_0000_0000_0000')


class SafetyNetworkNumber(unittest.TestCase):
    """Tests for safety network numbers."""
    def setUp(self):
        element = fixture.parse_xml(r"""<Module Name="Local" CatalogNumber="1756-L61S" Vendor="1" ProductType="14" ProductCode="67" Major="20" Minor="11" ParentModule="Local" ParentModPortId="1" Inhibited="false" MajorFault="true"
 SafetyNetwork="16#0000_4544_03d1_e91a">
<EKey State="ExactMatch"/>
<Ports>
<Port Id="1" Address="0" Type="ICP" Upstream="false" Width="2">
<Bus Size="10"/>
</Port>
</Ports>
</Module>""")
        self.module = module.Module(element)

    def test_invalid_snn_type(self):
        """Confirm setting SNN to a non-string raises an exception."""
        with self.assertRaises(TypeError):
            self.module.snn = 0

    def test_invalid_snn_value(self):
        """Confirm setting SSN to an out-of-range value raises an exception."""
        with self.assertRaises(ValueError):
            self.module.snn = '1000000000000'

    def test_invalid_snn_str(self):
        """Confirm setting SNN to a non-hex value raises an exception."""
        with self.assertRaises(ValueError):
            self.module.snn = 'not hex'

    def test_set_snn_underscore(self):
        """Test setting SNN to a value including underscores."""
        self.module.snn = '0000_1111_2222'
        self.assertEqual(self.module.element.attrib['SafetyNetwork'],
                         '16#0000_0000_1111_2222')



"""
Internal XML DOM helper inteface objects.
"""

import xml.etree.ElementTree as ElementTree


# Logix uses CDATA sections to enclose certain content, such as rung
# comments and tag descriptions. This inappropriate use of CDATA is
# difficult to maintain when parsing and generating XML as CDATA is not
# intended to be used as semantic markup. Without special handling only
# possible with the DOM API, these CDATA sections are merged into the
# surrounding text, which then does not import correctly into Logix.
# To simplify the application and allow use of the ElementTree API,
# CDATA sections are converted into normal elements before parsing,
# then back to CDATA when the project is written. This is the tag name
# used do enclose the original CDATA content.
CDATA_TAG = 'CDATAContent'


class CDATAElement(object):
    """
    This class manages access to CDATA content contained within a dedicated
    XML element, such as for tag descriptions and comments. An example
    construct would be:

    <parent> <!-- Parent element must already exist. -->

      <!--
      This and its single CDATA child can either be preexisting or
      created by this class.
      -->
      <comment>
        <CDATAContent>Foo Bar</CDATAContent>
      </comment>

    </parent>
    """
    def __init__(self, element=None, parent=None, name=None, attributes={}):
        # Find the CDATA child if a target element was given.
        try:
            self.cdata_element = element.find(CDATA_TAG)

        # Create a new element and CDATA grandchild under the parent if no
        # element was given.
        except AttributeError:
            self.cdata_parent = ElementTree.SubElement(parent, name, attributes)
            self.cdata_element = ElementTree.SubElement(self.cdata_parent,
                                                        CDATA_TAG)

        else:
            self.cdata_parent = element

    def __str__(self):
        """Returns the current string content."""
        if self.cdata_element.text is None:
            return ''
        return self.cdata_element.text

    def set(self, s):
        """Sets the CDATA section content."""
        self.cdata_element.text = s


def get_localized_cdata(parent, language):
    """Retrieves a CDATA string from a parent element."""
    # Single-language projects contain text directly under the parent
    # element.
    if language is None:
        element = parent

    # Multi-language projects keep text for each language in child
    # elements identified by the Lang attribute.
    else:
        path = "*[@Lang='{0}']".format(language)
        element = parent.find(path)
        if element is None:
            return None

    return CDATAElement(element)


def modify_localized_cdata(parent, language, text):
    """Alters CDATA content under a parent element."""
    cdata = get_localized_cdata(parent, language)
    cdata.set(text)


def create_localized_cdata(parent, language, text):
    """Creates new CDATA content under a parent element."""
    # Single-language projects store the CDATA directly under the
    # parent element.
    if language is None:
        ElementTree.SubElement(parent, CDATA_TAG)
        cdata = CDATAElement(parent)

    # Multi-language projects keep the CDATA in localized child elements
    # identified with the Lang attribute.
    else:
        tag_name = ''.join(('Localized', parent.tag))
        attr = {'Lang':language}
        cdata = CDATAElement(parent=parent, name=tag_name, attributes=attr)

    cdata.set(text)


def remove_localized_cdata(grandparent, parent, language):
    """Removes an element containing CDATA content."""
    # For multi-language projects, remove only the child associated
    # with the given language.
    if language is not None:
        cdata = get_localized_cdata(parent, language)
        if cdata is not None:
            parent.remove(cdata.cdata_parent)

    # Remove the entire parent element for single-language projects,
    # or if no comments remain in other languages.
    if (language is None) or (len(parent) == 0):
        grandparent.remove(parent)


class ElementDescription(object):
    """Descriptor class for accessing a top-level Description element.

    Description elements contain a CDATA comment for some type of
    container element, such as a Tag or Module.
    """
    def __init__(self, follow=[]):
        """Store a list of elements which must preceed the description."""
        self.follow = follow

    def __get__(self, instance, owner=None):
        """Returns the current description string."""
        desc = instance.element.find('Description')
        if desc is not None:
            cdata = get_localized_cdata(desc, instance.lang)
            if cdata is not None:
                return str(cdata)

        return None

    def __set__(self, instance, value):
        """Modifies the description text."""
        # Set a new description if given a string value.
        if isinstance(value, str):
            # See if the given description should replace an existing one
            # or be created as an entirely new element.
            if self.__get__(instance) is not None:
                self.modify(instance, value)
            else:
                self.create(instance, value)

        # A value of None removes any existing description.
        elif value is None:
            self.remove(instance)

        else:
            raise TypeError('Description must be a string or None')

    def modify(self, instance, value):
        """Alters the content of an existing description."""
        desc = instance.element.find('Description')
        modify_localized_cdata(desc, instance.lang, value)

    def create(self, instance, value):
        """Creates a new description when one does not previously exist."""

        # The Description element directly contains the text content in
        # single-language projects.
        if instance.lang is None:
            desc = ElementTree.Element('Description')
            self.insert_description(instance, desc)

        # Multi-language projects use localized child elements under
        # Description for each language.
        else:
            # Locate the Description tag, or create a new one if necessary.
            desc = instance.element.find('Description')
            if desc is None:
                desc = ElementTree.Element('Description')
                self.insert_description(instance, desc)

        create_localized_cdata(desc, instance.lang, value)

    def insert_description(self, instance, desc):
        """
        Inserts the Description element as a child of the parent instance
        based on any elements that must come first.
        """
        dest_index = 0 # Default to the first child.

        # Iterate through the child elements looking for any that
        # need to preceed the description.
        i = 0
        for e in instance.element:
            if e.tag in self.follow:
                dest_index = i + 1
            i += 1

        instance.element.insert(dest_index, desc)

    def remove(self, instance):
        """Implements deleting a description."""
        desc = instance.element.find('Description')
        if desc is not None:
            remove_localized_cdata(instance.element, desc, instance.lang)


class AttributeDescriptor(object):
    """Generic descriptor class for accessing an XML element's attribute."""
    def __init__(self, name, read_only=False):
        self.name = name
        self.read_only = read_only

    def __get__(self, instance, owner=None):
        try:
            raw = instance.element.attrib[self.name]
        except KeyError:
            return None
        else:
            return self.from_xml(raw)

    def __set__(self, instance, value):
        if self.read_only is True:
            raise AttributeError('Attribute is read-only')
        new_value = self.to_xml(instance, value)
        if new_value is not None:
            instance.element.attrib[self.name] = new_value

        # Delete the attribute if value is None, ignoring the case if the
        # attribute didn't exist to begin with.
        else:
            try:
                del instance.element.attrib[self.name]
            except KeyError:
                pass

    def from_xml(self, value):
        """Default converter for reading attribute string.

        Can be overridden in subclasses to provide custom conversion.
        """
        return str(value)

    def to_xml(self, instance, value):
        """Default converter for writing attribute string.

        Subclasses may implement custom conversions from user values
        by overriding this method. Must return a string or None.
        """
        if (value is not None) and (not isinstance(value, str)):
            raise TypeError('Value must be a string')
        return value


class ElementDictNames(object):
    """Descriptor class to get a list of an ElementDict's members."""
    def __get__(self, instance, owner=None):
        return [instance.key_type(e.attrib[instance.key_attr])
                for e in instance.parent.iterfind('*')]

    def __set__(self, instance, owner=None):
        """Raises an exception upon an attempt to modify; this is read-only."""
        raise AttributeError('Read-only attribute.')


class ElementDict(object):
    """Container which provides access to a group of XML elements.

    Operates similar to a dictionary where a child element is referenced
    by index notation to find the child with the matching key attribute.
    Instead of returning the actual XML element, a new object is
    instantiated and returned which is used to handle access to the child's
    data.
    """
    names = ElementDictNames()

    def __init__(self, parent, key_attr, value_type, type_attr=None,
                 dfl_type=None, key_type=str, value_args=[]):
        self.parent = parent
        self.key_attr = key_attr
        self.value_type = value_type
        self.type_attr = type_attr
        self.dfl_type = dfl_type
        self.key_type = key_type
        self.value_args = value_args

    def __getitem__(self, key):
        """Return a member class suitable for accessing a child element."""
        xpath = "*[@{0}='{1}']".format(self.key_attr, key)
        element = self.parent.find(xpath)
        if element is None:
            raise KeyError("{0} not found".format(key))
        return self.create_value_object(element)

    def create_value_object(self, element):
        """Instantiates an object returned as the value."""
        args = [element]
        args.extend(self.value_args)

        # Use the same type for all child objects if the given value type
        # is a class.
        try:
            return self.value_type(*args)

        # If the value type is a mapping, use the given attribute to select
        # the appropriate type.
        except TypeError:
            type_name = element.attrib[self.type_attr]
            return self.value_type.get(type_name, self.dfl_type)(*args)



"""
Unit tests for a project's programs object.
"""

from tests import fixture
import unittest


class Programs(unittest.TestCase):
    """Tests for the top-level programs container object."""
    def setUp(self):
        prj = fixture.string_to_project(r"""<?xml version="1.0" encoding="UTF-8" standalone="yes"?>
<RSLogix5000Content SchemaRevision="1.0" SoftwareRevision="20.01" TargetName="test" TargetType="Controller" ContainsContext="false" Owner="admin" ExportDate="Mon Jul 20 01:45:55 2020" ExportOptions="DecoratedData ForceProtectedEncoding AllProjDocTrans">
<Controller Use="Target" Name="test" ProcessorType="1756-L61" MajorRev="20" MinorRev="11" TimeSlice="20" ShareUnusedTimeSlice="1" ProjectCreationDate="Sat Jul 18 23:53:16 2020" LastModifiedDate="Sat Jul 18 23:53:18 2020" SFCExecutionControl="CurrentActive" SFCRestartPosition="MostRecent"
 SFCLastScan="DontScan" ProjectSN="16#0000_0000" MatchProjectToController="false" CanUseRPIFromProducer="false" InhibitAutomaticFirmwareUpdate="0">
<Programs>
<Program Name="MainProgram" TestEdits="false" Disabled="false">
<Tags>
<Tag Name="main_tag_1" TagType="Base" DataType="DINT" Radix="Decimal" Constant="false" ExternalAccess="Read/Write">
<Data>00 00 00 00</Data>
<Data Format="Decorated">
<DataValue DataType="DINT" Radix="Decimal" Value="0"/>
</Data>
</Tag>
<Tag Name="main_tag_2" TagType="Base" DataType="DINT" Radix="Decimal" Constant="false" ExternalAccess="Read/Write">
<Data>00 00 00 00</Data>
<Data Format="Decorated">
<DataValue DataType="DINT" Radix="Decimal" Value="0"/>
</Data>
</Tag>
</Tags>
<Routines/>
</Program>
<Program Name="prog2" TestEdits="false" Disabled="false">
<Tags>
<Tag Name="prog2_tag_1" TagType="Base" DataType="DINT" Radix="Decimal" Constant="false" ExternalAccess="Read/Write">
<Data>00 00 00 00</Data>
<Data Format="Decorated">
<DataValue DataType="DINT" Radix="Decimal" Value="0"/>
</Data>
</Tag>
<Tag Name="prog2_tag_2" TagType="Base" DataType="DINT" Radix="Decimal" Constant="false" ExternalAccess="Read/Write">
<Data>00 00 00 00</Data>
<Data Format="Decorated">
<DataValue DataType="DINT" Radix="Decimal" Value="0"/>
</Data>
</Tag>
</Tags>
<Routines/>
</Program>
</Programs>
</Controller>
</RSLogix5000Content>""")
        self.programs = prj.programs

    def test_names_read(self):
        """Test name attribute returns all program names."""
        self.assertEqual(set(self.programs.names),
                         set(('MainProgram', 'prog2')))

    def test_names_read_only(self):
        """Ensure names attribute is read-only."""
        with self.assertRaises(AttributeError):
            self.programs.names = 'foo'

    def test_index(self):
        """Test indexing by names."""
        for prg in self.programs.names:
            self.programs[prg]

    def test_tags_names(self):
        """Ensure tags names attribute is a iterable of non-empty strings."""
        self.assertEqual(set(self.programs['MainProgram'].tags.names),
                         set(('main_tag_1', 'main_tag_2')))
        self.assertEqual(set(self.programs['prog2'].tags.names),
                         set(('prog2_tag_1', 'prog2_tag_2')))



"""
Top-level module defining the Project class through which all L5X access
is performed.

The general approach of this package is to provide access to L5X data
through descriptor objects organized under a Project instance in a structure
similar to RSLogix. These descriptor objects modify XML elements through
the __set__() method or return appropriately converted data from the
__get__() method. In this way the application can process L5X projects
without worrying about low-level XML handling.
"""

from .dom import (CDATA_TAG, ElementDict, AttributeDescriptor)
from .module import (Module, SafetyNetworkNumber)
from .tag import Scope
import io
import re
import xml.etree.ElementTree as ElementTree
import xml.dom.minidom


class InvalidFile(Exception):
    """Raised if the given .L5X file was not a proper L5X export."""
    pass


class Project(object):
    """Top-level container for an entire Logix project."""
    def __init__(self, filename):
        # Dummy document used only for generating replacement CDATA sections.
        implementation = xml.dom.minidom.getDOMImplementation()
        self.cdata_converter = implementation.createDocument(None, None, None)

        self.parse(filename)

        # Confirm the root element indicates this is a Logix project.
        if self.doc.tag != 'RSLogix5000Content':
            raise InvalidFile('Not an L5X file.')

        try:
            lang = self.doc.attrib['CurrentLanguage']
        except KeyError:
            lang = None

        ctl_element = self.doc.find('Controller')
        self.controller = Controller(ctl_element, lang)

        progs = ctl_element.find('Programs')
        self.programs = ElementDict(progs, 'Name', Scope, value_args=[lang])

        mods = ctl_element.find('Modules')
        self.modules = ElementDict(mods, 'Name', Module)

    def parse(self, filename):
        """Parses the source project."""
        # Accept both filename strings for normal usage, and buffer objects
        # for unit tests.
        try:
            f = io.open(filename, encoding='UTF-8')
        except TypeError:
            f = filename

        with f:
            orig = f.read()

        # Swap out CDATA sections before parsing.
        cdata_replaced = self.convert_to_cdata_element(orig)

        # The (unicode) string needs to be converted back to a series
        # of bytes before ElementTree parses it.
        encoded = cdata_replaced.encode('UTF-8')

        try:
            self.doc = ElementTree.fromstring(encoded)
        except ElementTree.ParseError as e:
            raise InvalidFile("XML parsing error: {0}".format(e))

    def convert_to_cdata_element(self, doc):
        """Replaces the delimiters surrounding CDATA sections.

        This is used before parsing to convert CDATA sections into
        normal elements.
        """
        pattern = r"""
            <!\[CDATA\[   # Opening CDATA sequence.
            (?P<text>.*?) # Element content.
            \]\]>         # Closing CDATA sequence.
        """
        return re.sub(pattern, self.cdata_element, doc,
                      flags=re.VERBOSE | re.DOTALL)

    def cdata_element(self, match):
        """
        Generates a string representation of an XML element with a given
        text content. Used when replacing CDATA sections with elements.
        """
        element = ElementTree.Element(CDATA_TAG)
        element.text = match.group('text')
        return ElementTree.tostring(element).decode()

    def convert_to_cdata_section(self, doc):
        """Replaces CDATA elements with CDATA sections.

        This is used after writing a project to reinstall the CDATA sections
        required by RSLogix.
        """
        # Regular expression pattern to locate CDATA elements.
        pattern = r"""
            # Match elements with separate opening and closing tags.
            <{0}\s*>  # Opening tag.
            .*?       # Element content.
            </{0}\s*> # Closing tag

            |

            # Also match empty, self-closing tags.
            <{0}\s*/>
        """.format(CDATA_TAG)
        return re.sub(pattern, self.cdata_section, doc,
                      flags=re.VERBOSE | re.DOTALL)

    def cdata_section(self, match):
        """
        Generates a string representation of a CDATA section containing
        the content of an XML element. Used when replacing elements with
        CDATA sections.
        """
        # Parse the string to handle any escape sequences or character
        # references, after encoding to ensure ElementTree parses a
        # series of bytes.
        encoded = match.group().encode('UTF-8')
        root = ElementTree.fromstring(encoded)

        # Empty elements have None for text, which cannot be used when
        # creating a CDATA section.
        if root.text is not None:
            text = root.text
        else:
            text = ''

        cdata = self.cdata_converter.createCDATASection(text)
        return cdata.toxml()

    def write(self, filename):
        """Outputs the document to a new file."""
        no_cdata = self.doc_to_string()
        with_cdata = self.convert_to_cdata_section(no_cdata)

        encoded = with_cdata.encode('UTF-8')
        try:
            f = open(filename, 'wb')

        # Accept buffer targets for unit testing.
        except TypeError:
            filename.write(encoded)
            filename.seek(0)

        else:
            with f:
                f.write(encoded)

    def doc_to_string(self):
        """Serializes the document into a unicode string.

        This is used instead of ElementTree's tostring() method because
        an XML declaration is required, which can only be generated
        with write().
        """
        tree = ElementTree.ElementTree(self.doc)
        buf = io.BytesIO()
        tree.write(buf, encoding='UTF-8', xml_declaration=True)
        return buf.getvalue().decode('UTF-8')


class Controller(Scope):
    """Accessor object for the controller device."""
    comm_path = AttributeDescriptor('CommPath')

    # The safety network number is stored in the first module element,
    # not the Controller element.
    snn = SafetyNetworkNumber('Modules/Module')



"""
Utilities used by all test modules to ensure test output is accumulated into
a single output file for final validation by RSLogix.
"""

import io
import l5x
import xml.dom.minidom
import xml.etree.ElementTree as ElementTree


def parse_xml(xml_str):
    """Parses XML from a string."""
    class Parser(l5x.Project):
        """Use the Project class to handle CDATA conversion."""
        def __init__(self, s):
            # Swap out CDATA sections before parsing.
            cdata_replaced = self.convert_to_cdata_element(s)

            # The (unicode) string needs to be converted back to a series
            # of bytes before ElementTree parses it.
            encoded = cdata_replaced.encode('UTF-8')

            self.doc = ElementTree.fromstring(encoded)

    parser = Parser(xml_str)
    return parser.doc


def string_to_project(s):
    """Parses an XML string into a L5X project."""
    # Convert to unicode as needed for Python 2.7.
    try:
        s = unicode(s)
    except NameError:
        pass

    buf = io.StringIO(s)
    return l5x.Project(buf)


def create_project(*populate):
    """
    Generates a mock L5X project. The created document contains only an
    empty set of elements required by the L5X Project class; the populate
    callbacks are used to add whatever additional content needed by particular
    test cases.
    """
    imp = xml.dom.minidom.getDOMImplementation()
    doc = imp.createDocument(None, 'RSLogix5000Content', None)
    root = doc.documentElement

    controller = doc.createElement('Controller')
    root.appendChild(controller)

    # Create the top-level elements under the Controller.
    for tag in ['Tags', 'Programs', 'Modules']:
        element = doc.createElement(tag)
        controller.appendChild(element)

    # Dispatch the document to the populate callbacks to allow additional
    # content to be added before serialization.
    [f(doc) for f in populate]

    # Serialize the document so it can be parsed as a simulated XML file.
    xml_str = root.toxml('UTF-8').decode('UTF-8')
    buf = io.StringIO(xml_str)
    return l5x.Project(buf)



"""
Objects for accessing a set of I/O modules.
"""

from .dom import (ElementDict, AttributeDescriptor)


class SafetyNetworkNumber(object):
    """Descriptor class for accessing safety network numbers."""
    ATTRIBUTE_NAME = 'SafetyNetwork'
    PREFIX = '16#0000'

    def __init__(self, element_path='.'):
        self.element_path = element_path

    def __get__(self, instance, owner=None):
        """Returns the current SNN.

        Removes the prefix, unused 16 most-significant bits, and underscores.
        """
        element = self.get_target_element(instance)
        self.check_is_safety(element)
        snn = element.attrib[self.ATTRIBUTE_NAME][len(self.PREFIX):]
        return snn.replace('_', '')

    def __set__(self, instance, value):
        """Sets a new SNN."""
        element = self.get_target_element(instance)
        self.check_is_safety(element)
        if not isinstance(value, str):
            raise TypeError('Safety network number must be a hex string')
        new = value.replace('_', '')

        # Ensure valid hex string.
        try:
            x = int(new, 16)
        except ValueError:
            raise ValueError('Safety network number must be a hex string')

        # Generate a zero-padded string and enforce 24-bit limit.
        padded = "{0:012X}".format(x)
        if not len(padded) == 12:
            raise ValueError('Value must be 24-bit, 12 hex characters')

        # Add radix prefix and insert underscores for the final output string.
        fields = [self.PREFIX]
        for word in range(3):
            start = word * 4
            end = start + 4
            fields.append(padded[start:end])

        element.attrib[self.ATTRIBUTE_NAME] = '_'.join(fields)

    def get_target_element(self, instance):
        """Finds the element containing the safety network number attribute."""
        return instance.element.find(self.element_path)

    def check_is_safety(self, element):
        """Confirms the target port/module is safety and has a SNN."""
        try:
            element.attrib[self.ATTRIBUTE_NAME]
        except KeyError:
            try:
                id = element.attrib['Name']
            except KeyError:
                id = "{0}({1})".format(element.attrib['Id'],
                                       element.attrib['Type'])
            msg = "{0} {1} does not support a safety network number.".format(
                element.tag, id)
            raise TypeError(msg)


class NatAddress(AttributeDescriptor):
    """Descriptor class for accessing port NAT addresses."""
    def to_xml(self, port, new_address):
        """
        Override for the default converter method to enforce custom validations
        specifc to NAT addresses.
        """
        # The port must already be configured for NAT, i.e. the NAT XML
        # attribute must already exist.
        if not self.name in port.element.attrib:
            raise TypeError("Port {0}({1}) is not configured for NAT.".format(
                port.element.attrib["Id"], port.element.attrib["Type"]))

        # Disallow None as removing the NAT address is not permitted.
        if not isinstance(new_address, str):
            raise TypeError("NAT address must be a string.")

        return new_address


class Module(object):
    """Accessor object for a communication module."""
    snn = SafetyNetworkNumber()

    def __init__(self, element):
        self.element = element
        ports_element = element.find('Ports')
        self.ports = ElementDict(ports_element, 'Id', Port, key_type=int)


class Port(object):
    """Accessor object for a module's port."""
    address = AttributeDescriptor('Address')
    nat_address = NatAddress('NATActualAddress')
    type = AttributeDescriptor('Type', True)
    snn = SafetyNetworkNumber()

    def __init__(self, element):
        self.element = element



"""
Objects implementing tag access.
"""

from l5x import dom
import copy
import ctypes
import itertools
import xml.etree.ElementTree as ElementTree


class Scope(object):
    """Container to hold a group of tags within a specific scope."""
    def __init__(self, element, lang):
        self.element = element
        tag_element = element.find('Tags')
        self.tags = dom.ElementDict(tag_element, 'Name', Tag,
                                    value_args=[lang])


class TagDataDescriptor(object):
    """Descriptor class to dispatch attribute access to a data object.

    Used by Tag objects to pass access to a specific attribute on to the
    Data element which handles the implementation.
    """
    def __init__(self, attr):
        self.attr = attr

    def __get__(self, tag, owner=None):
        return getattr(tag.data, self.attr)

    def __set__(self, tag, value):
        setattr(tag.data, self.attr, value)


class ConsumeDescriptor(object):
    """Descriptor class for accessing consumed tag properties."""
    def __init__(self, attr):
        self.attr = attr

    def __get__(self, tag, owner=None):
        """Returns the current consumed tag property."""
        self.check_consumed(tag)
        info = self.get_info(tag)
        return info.attrib[self.attr]
        
    def __set__(self, tag, value):
        """Sets a new consumed tag property."""
        self.check_consumed(tag)
        
        # Producer names must be non-empty strings.
        if not isinstance(value, str):
            raise TypeError('Producer must be a string')
        if len(value) == 0:
            raise ValueError('Producer string cannot be empty')

        info = self.get_info(tag)
        info.attrib[self.attr] = value

    def check_consumed(self, tag):
        """Verifies this is a consumed tag."""
        if tag.element.attrib['TagType'] != 'Consumed':
            raise TypeError("Tag {0} is not a consumed tag".format(
                tag.element.attrib['Name']))

    def get_info(self, tag):
        """Retrieves the ConsumeInfo XML element."""
        return tag.element.find('ConsumeInfo')


class Tag(object):
    """Base class for a single tag."""
    description = dom.ElementDescription(['ConsumeInfo'])
    data_type = dom.AttributeDescriptor('DataType', True)
    value = TagDataDescriptor('value')
    shape = TagDataDescriptor('shape')
    names = TagDataDescriptor('names')
    producer = ConsumeDescriptor('Producer')
    remote_tag = ConsumeDescriptor('RemoteTag')

    def __new__(cls, element, lang):
        """
        Intercepts the creation of a new tag object to determine if
        the target tag is an alias, in which case an alias tag object
        is returned instead of a Tag.
        """
        if element.attrib['TagType'] == 'Alias':
            alias = object.__new__(AliasTag)
            alias.__init__(element, lang)
            return alias

        # Normal base tag; return an instance of this class.
        return object.__new__(cls)

    def __init__(self, element, lang):
        self.element = element
        self.lang = lang
        data_class = base_data_types.get(self.data_type, Structure)
        self.data = data_class(self.get_data_element(), self)

    def get_data_element(self):
        """Returns the decorated data XML element.

        This is always the sole element contained with the decorated Data
        element.
        """
        data = self.element.find("Data[@Format='Decorated']/*")

        if data is None:
            name = self.element.attrib['Name']
            raise RuntimeError("Decoded data content not found for {0} tag. "
                               "Ensure Encode Source Protected Content option "
                               "is disabled when saving L5X.".format(name))

        return data

    def __getitem__(self, key):
        """
        Indices are passed to the data object to access members of compound
        data types.
        """
        return self.data[key]

    def __len__(self):
        """Dispatches len queries to the base data type object."""
        return len(self.data)

    def clear_raw_data(self):
        """Removes any data elements other than decorated.
        
        Called anytime a data value is set to avoid conflicts with
        modified decorated data elements.
        """
        undecorated_data = []
        for e in self.element.iterfind('Data'):
            try:
                format = e.attrib['Format']
            except KeyError:
                undecorated_data.append(e)
            else:
                if format != 'Decorated':
                    undecorated_data.append(e)

        [self.element.remove(e) for e in undecorated_data]


class AliasFor(object):
    """Descriptor class to access the AliasFor attribute."""
    def __get__(self, tag, owner=None):
        return tag.element.attrib['AliasFor']

    def __set__(self, tag, value):
        if not isinstance(value, str):
            raise TypeError('Alias tag name must be a string.')
        if not value.strip():
            raise ValueError('Alias tag name must be a non-empty string.')
        tag.element.attrib['AliasFor'] = value
        self.remove_operand_comments(tag)

    def remove_operand_comments(self, tag):
        """Deletes any comments for the tag's operands.

        This is done because the new alias target data type is unknown,
        and comments from previous operands may not be valid for the
        new tag.
        """
        comments = tag.element.find('Comments')
        if comments is not None:
            tag.element.remove(comments)


class AliasTag(object):
    """Handler for accessing alias tags."""
    description = dom.ElementDescription()
    alias_for = AliasFor()

    def __init__(self, element, lang):
        self.element = element
        self.lang = lang


class Comment(object):
    """Descriptor class for accessing descriptions of individual tag members.

    These descriptions are stored in the Comments element directly under
    the enclosing Tag element. The instance's operand attribute is used
    to find the correct Comment element.
    """
    def __get__(self, instance, owner=None):
        """Returns the data's description."""
        # Acquire the overall Comments parent element.
        comments = instance.tag.element.find('Comments')
        if comments is None:
            return None

        # Locate the Comment child with the matching operand.
        try:
            element = self.get_comment_element(instance, comments)
        except KeyError:
            return None

        cdata = dom.get_localized_cdata(element, instance.tag.lang)
        if cdata is None:
            return None
        return str(cdata)

    def __set__(self, instance, value):
        """Updates, creates, or removes a comment."""
        if value is not None:
            if self.__get__(instance) is None:
                self.create(instance, value)
            else:
                self.modify(instance, value)
        else:
            self.delete(instance)

    def create(self, instance, text):
        """Creates a new comment."""
        # Get the parent Comments element, creating one if necessary.
        comments = instance.tag.element.find('Comments')
        if comments is None:
            comments = self.create_comments(instance)

        # Find or create a Comment element with matching operand to store
        # the new comment text.
        try:
            # Single-language projects will not have an existing Comment
            # element because no localized comments are possible in other
            # languages.
            if instance.tag.lang is None:
                raise KeyError()

            # A matching Comment element may already exist in multilanguage
            # projects, containing comments in other languages.
            else:
                comment = self.get_comment_element(instance, comments)

        # Create a new Comment element with the target operand.
        except KeyError:
            comment = ElementTree.SubElement(comments, 'Comment',
                                             {'Operand':instance.operand})

        dom.create_localized_cdata(comment, instance.tag.lang, text)

    def modify(self, instance, text):
        """Alters an existing comment."""
        comments_parent = instance.tag.element.find('Comments')
        comment = self.get_comment_element(instance, comments_parent)
        dom.modify_localized_cdata(comment, instance.tag.lang, text)

    def delete(self, instance):
        """Removes a comment."""
        # Acquire the overall Comments parent element.
        comments = instance.tag.element.find('Comments')
        if comments is None:
            return

        # Locate the Comment child with the matching operand.
        try:
            comment = self.get_comment_element(instance, comments)
        except KeyError:
            return

        # Remove the Comment or LocalizedComment containing the actual text.
        dom.remove_localized_cdata(comments, comment, instance.tag.lang)

        # Remove the entire Comments parent element if no other comments for any
        # operands remain.
        if len(comments) == 0:
            instance.tag.element.remove(comments)

    def create_comments(self, instance):
        """Creates a new Comments container element.

        Used if the top-level tag element did not contain a Comments element.
        The Comments element must be located immediately before any Data
        elements.
        """
        comments = ElementTree.Element('Comments')

        # Locate the index of the Data child element.
        child_tags = [e.tag for e in instance.tag.element.iterfind('*')]
        data_index = child_tags.index('Data')

        instance.tag.element.insert(data_index, comments)
        return comments

    def get_comment_element(self, instance, comments):
        """Acquires the Comment element of the instance's operand."""
        path = "Comment[@Operand='{0}']".format(instance.operand)
        element = comments.find(path)
        if element is None:
            raise KeyError()
        return element


class Data(object):
    """Base class for objects providing access to tag values and comments."""
    description = Comment()

    def __new__(cls, *args, **kwds):
        """
        Intercepts creation of a new data object if the XML element
        indicates it is an array, in which case an array access object
        is created instead for the given data type.
        """
        if args[0].tag.startswith('Array'):

            # Two array accessor types are possible depending on if the
            # the array is a structure member.
            if args[0].tag == ('ArrayMember'):
                array_type = ArrayMember
            else:
                array_type = Array

            array = object.__new__(array_type)
            array_args = [cls]
            array_args.extend(args)
            array.__init__(*array_args, **kwds)
            return array

        # Non-array tags return a instance of the original type; an explicit
        # call to __init__ is not required as the returned instance
        # is the original class.
        else:
            return object.__new__(cls)

    def __init__(self, element, tag, parent=None):
        self.element = element
        self.tag = tag
        self.parent = parent
        self.build_operand()

    def build_operand(self):
        """Constructs the identifier for comment operands.

        A tag's top-level data type has no parent and does not require
        an operand; it's description is placed in the dedicated Description
        element. These objects get an empty operand string for child
        members to use.

        Operands for sub-members are formed by appending their name
        to their parent's operand. Names are converted to upper-case
        because Logix uses only capital letters in operand attributes
        for some reason.
        """
        if self.parent is None:
            self.operand = ''
        else:
            # One of two possible XML attributes determine how this data is
            # identified.
            try:
                # Array members use the Index attribute, which is appended to
                # the operand string without an additional separator; the
                # enclosing square brackets are included in the attribute
                # value.
                operand = self.element.attrib['Index']
                sep = ''
            except KeyError:
                # All other operands use the Name attribute, which is
                # separated from other operands by a dot.
                operand = self.element.attrib['Name']
                sep = '.'
                
            self.operand = sep.join((self.parent.operand, operand.upper()))


class IntegerValue(object):
    """Descriptor class for accessing an integer's value."""
    def __get__(self, instance, owner=None):
        return int(instance.element.attrib['Value'])

    def __set__(self, instance, value):
        """Sets a new value."""
        if (not isinstance(value, int)) or isinstance(value, bool):
            raise TypeError('Value must be an integer')
        if (value < instance.value_min) or (value > instance.value_max):
            raise ValueError('Value out of range')
        instance.element.attrib['Value'] = str(value)
        instance.tag.clear_raw_data()


class Integer(Data):
    """Base class for integer data types.

    In addition to the usual value and description access, integer indices
    are used for bit-level references.
    """
    value = IntegerValue()

    def __getitem__(self, bit):
        """Gets an object to access a single bit."""
        self.validate_bit_number(bit)
        return Bit(self.element, self.tag, self, bit)

    def validate_bit_number(self, bit):
        """Verifies a given bit index is within range."""
        if not isinstance(bit, int):
            raise TypeError('Bit indices must be integers.')
        if (bit < 0) or (bit >= self.bits):
            raise IndexError('Bit index out of range')

    def __len__(self):
        """Returns the width of the integer."""
        return self.bits


class SINT(Integer):
    """Base class for 8-bit signed integers."""
    bits = 8
    ctype = ctypes.c_int8
    value_min = -128
    value_max = 127


class INT(Integer):
    """Base class for 16-bit signed integers."""
    bits = 16
    ctype = ctypes.c_int16
    value_min = -32768
    value_max = 32767


class DINT(Integer):
    """Base class for 32-bit signed integers."""
    bits = 32
    ctype = ctypes.c_int32
    value_min = -2147483648
    value_max = 2147483647


class BitValue(object):
    """Descriptor class for values of individual integer bits.

    Bit access utilizes exact-width, signed ctype integers for
    bit-level operations which are then translated back to the parent
    integer's value. This ensures correct results when the sign bit
    is accessed.
    """
    def __get__(self, bit, owner=None):
        cvalue = self.get_ctype(bit)
        if cvalue.value & bit.mask.value:
            return 1
        else:
            return 0

    def __set__(self, bit, bit_value):
        if not isinstance(bit_value, int):
            raise TypeError('Bit values must be integers')
        elif (bit_value < 0) or (bit_value > 1):
            raise ValueError('Bit values may only be 0 or 1')

        cvalue = self.get_ctype(bit)
        if bit_value:
            cvalue.value |= bit.mask.value
        else:
            cvalue.value &= ~bit.mask.value
        bit.parent.value = int(cvalue.value)

    def get_ctype(self, bit):
        """Returns the parent integer's value as a ctype."""
        return bit.parent.ctype(bit.parent.value)


class Bit(Data):
    """Provides access to individual bits within an integer."""
    value = BitValue()
    description = Comment()

    def __init__(self, element, tag, parent, bit):
        self.bit = bit
        Data.__init__(self, element, tag, parent)
        self.mask = parent.ctype(1 << bit)

    def build_operand(self):
        """Method override to create an operand based on the bit number."""
        self.operand = '.'.join((self.parent.operand, str(self.bit)))


class BOOL(Data):
    """Tag access for BOOL data types."""
    value = IntegerValue()
    value_min = 0
    value_max = 1


class RealValue(object):
    """Descriptor class for accessing REAL values."""
    def __get__(self, instance, owner=None):
        return float(instance.element.attrib['Value'])

    def __set__(self, instance, value):
        if not isinstance(value, float):
            raise TypeError('Value must be a float')

        # Check for NaN and infinite values.
        try:
            value.as_integer_ratio()
        except (OverflowError, ValueError):
            raise ValueError('NaN and infinite values are not supported')
            
        instance.element.attrib['Value'] = str(value)
        instance.tag.clear_raw_data()


class REAL(Data):
    """Tag access for REAL data types."""
    value = RealValue()


class StructureValue(object):
    """Descriptor class for accessing multiple structure values.

    Values are expressed as a dictionary with member names as keys.
    """
    def __get__(self, struct, owner=None):
        member_names = struct.members.names
        return dict(zip(member_names, [struct[m].value for m in member_names]))

    def __set__(self, struct, value):
        if not isinstance(value, dict):
            raise TypeError('Value must be a dictionary')
        for m in value.keys():
            struct[m].value = value[m]
        struct.tag.clear_raw_data()


class StructureNames(object):
    """Descriptor class for accessing structure member names."""
    def __get__(self, struct, owner=None):
        return struct.members.names

    def __set__(self, struct, owner=None):
        raise AttributeError('Read-only attribute.')


class Structure(Data):
    """Accessor class for structured data types."""
    value = StructureValue()
    names = StructureNames()

    def __init__(self, element, tag, parent=None):
        Data.__init__(self, element, tag, parent)

        # If this structure is an array member the given XML element
        # is just the enclosing array member; the XML element directly
        # holding the structure's data is the first child: a Structure
        # XML element.
        if element.tag == 'Element':
            self.element = element.find('Structure')

        self.members = dom.ElementDict(self.element, 'Name', base_data_types,
                                   'DataType', Structure,
                                   value_args=[tag, self])

    def __getitem__(self, member):
        """Indexing a structure yields an individual member."""
        if not isinstance(member, str):
            raise TypeError('Structure indices must be strings')
        return self.members[member]


class ArrayValue(object):
    """Descriptor class for accessing multiple values in an array."""
    def __get__(self, array, owner=None):
        dim = len(array.shape) - len(array.address) - 1
        return [array[i].value for i in range(array.shape[dim])]

    def __set__(self, array, value):
        if not isinstance(value, list):
            raise TypeError('Value must be a list')
        if len(value) > array.shape[len(array.shape) - len(array.address) - 1]:
            raise IndexError('Source list is too large')

        for i in range(len(value)):
            array[i].value = value[i]

        array.tag.clear_raw_data()


class ArrayDescription(Comment):
    """Descriptor class array descriptions.

    Raises an exception for an attempts to access descriptions because
    RSLogix does not support commenting subarrays; only individual elements
    may have descriptions.
    """
    e = TypeError
    msg = 'Descriptions for subarrays are not supported'

    def __get__(self, array, owner=None):
        raise self.e(self.msg)

    def __set__(self, array, value):
        raise self.e(self.msg)


class ArrayShape(object):
    """Descriptor class to acquire an array's dimensions."""
    def __get__(self, array, owner=None):
        dims = [int(d) for d in array.element.attrib['Dimensions'].split(',')]

        # Dimensions are stored most-significant first(Dim2, Dim1, Dim0) in the
        # XML attribute; reversing them makes DimX = shape[X].
        dims.reverse()

        return tuple(dims)

    def __set__(self, array, value):
        # Prevent resizing UDT array members.
        if not array.element.tag == 'Array':
            raise AttributeError('Member arrays cannot be resized.')

        self.check_shape(value)
        array.resize(value)

    def check_shape(self, shape):
        """Validates a new target shape before resizing."""
        if not isinstance(shape, tuple):
            raise TypeError('Array shape must be a tuple')

        dims = len(shape)
        if (dims < 1) or (dims > 3):
            raise ValueError('Arrays must have between 1 and 3 dimensions')

        for d in shape:
            if not isinstance(d, int):
                raise TypeError('Array dimensions must be integers')

            if d < 1:
                raise ValueError('Array dimension must be >= 1')


class Array(Data):
    """Access object for arrays of any data type."""
    value = ArrayValue()
    description = ArrayDescription()
    shape = ArrayShape()

    def __init__(self, data_class, element, tag, parent=None, address=[]):
        Data.__init__(self, element, tag, parent)
        self.data_class = data_class
        self.address = address

        # Array members are identified in XML by the Index attribute,
        # not element order, and may include more than one dimension, so
        # the usual list-type access does not suffice. The object initialized
        # here builds a dictionary of child elements(array members) keyed
        # by the Index attribute that can then be accessed with traditional
        # array notation.
        self.members = dom.ElementDict(self.element, 'Index', self.data_class,
                                       value_args=[self.tag, self])

    def __getitem__(self, index):
        """Returns an access object for the given index.

        Multidimensional arrays will return new Array objects with the
        accumulated address until all dimensions are satisfied, which
        will then return the data access object for that item.
        """
        if not isinstance(index, int):
            raise TypeError('Array indices must be integers')

        # Add the given index to the current accumulated address.
        dim = len(self.shape) - len(self.address) - 1
        if (index < 0) or (index >= self.shape[dim]):
            raise IndexError('Array index out of range')
        new_address = list(self.address)
        new_address.insert(0, index)

        # If the newly formed address set satisifies all dimensions
        # return an access object for the member.
        if len(new_address) == len(self.shape):
            # Address values are reversed because the display order is
            # most-significant first.
            new_address.reverse()

            key = "[{0}]".format(','.join([str(i) for i in new_address]))
            return self.members[key]

        # The new address does not yet specify a single element if the key
        # was not found. Return a new array access object to handle
        # access to the new address by instantiating the data type,
        # which will result in an Array instance through Data.__new__().
        else:
            return self.data_class(self.element, self.tag, self.parent,
                                   new_address)

    def resize(self, new_shape):
        """Alters the array's size."""
        self.set_dimensions(new_shape)
        self.tag.clear_raw_data()

        # Make a copy of the first element before stripping all old values.
        template = self.element.find('Element')

        self.remove_elements()

        # Generate new elements based on a new set of indices.
        indices = self.build_new_indices(new_shape)
        [self.append_element(template, i) for i in indices]

    def set_dimensions(self, shape):
        """Updates the Dimensions attributes with a given shape.

        Array tag elements have two dimension attributes: one in the top-level
        Tag element, and another in the Array child element.
        """
        new = list([str(x) for x in shape])
        new.reverse() # Logix lists dimensions most-significant first.

        # Top-level Tag element uses space for separators.
        value = ' '.join(new)
        self.tag.element.attrib['Dimensions'] = value

        # Array element uses comma for separators.
        value = ','.join(new)
        self.element.attrib['Dimensions'] = value

    def remove_elements(self):
        """Deletes all (array)Element elements."""
        [self.element.remove(e) for e in self.element.findall('*')]

    def build_new_indices(self, shape):
        """Constructs a set of all indices for a given array shape."""
        indices = [range(x) for x in shape]
        indices.reverse() # Indices are listed most-significant first.
        return itertools.product(*indices)

    def append_element(self, template, index):
        """Generates and appends a new element from a template."""
        new = copy.deepcopy(template)
        index_attr = "[{0}]".format(','.join([str(i) for i in index]))
        new.attrib['Index'] = index_attr
        self.element.append(new)


class ArrayMember(Array):
    """Access object for arrays which are structure members.

    Permits access to a description for the entire member. Preventing
    comments for subarrays is unnecessary as array members may only be
    one-dimensional.
    """
    description = Comment()

    
base_data_types = {'SINT':SINT,
                   'INT':INT,
                   'DINT':DINT,
                   'BOOL':BOOL,
                   'REAL':REAL}



"""
Unittests for tag access.
"""

import copy
import ctypes
from tests import fixture
from l5x import (dom, tag)
import itertools
import l5x
import math
import unittest
import xml.etree.ElementTree as ElementTree


class Scope(unittest.TestCase):
    """Tests for a tag scope."""
    def setUp(self):
        e = fixture.parse_xml("""<Program Name="MainProgram" TestEdits="false" MainRoutineName="MainRoutine" Disabled="false">
<Tags>
<Tag Name="bar" TagType="Base" DataType="DINT" Radix="Decimal" Constant="false" ExternalAccess="Read/Write">
<Data>00 00 00 00</Data>
<Data Format="Decorated">
<DataValue DataType="DINT" Radix="Decimal" Value="0"/>
</Data>
</Tag>
<Tag Name="baz" TagType="Base" DataType="DINT" Radix="Decimal" Constant="false" ExternalAccess="Read/Write">
<Data>00 00 00 00</Data>
<Data Format="Decorated">
<DataValue DataType="DINT" Radix="Decimal" Value="0"/>
</Data>
</Tag>
<Tag Name="foo" TagType="Base" DataType="DINT" Radix="Decimal" Constant="false" ExternalAccess="Read/Write">
<Data>00 00 00 00</Data>
<Data Format="Decorated">
<DataValue DataType="DINT" Radix="Decimal" Value="0"/>
</Data>
</Tag>
</Tags>
<Routines>
<Routine Name="MainRoutine" Type="RLL"/>
</Routines>
</Program>""")
        self.scope = l5x.tag.Scope(e, None)

    def test_names(self):
        """Test names attribute returns a non-empty list of strings."""
        self.assertGreater(len(self.scope.tags.names), 0)
        for tag in self.scope.tags.names:
            self.assertIsInstance(tag, str)
            self.assertGreater(len(tag), 0)

    def test_name_index(self):
        """Ensure tags can be indexed by name."""
        for name in self.scope.tags.names:
            self.scope.tags[name]

    def test_name_read_only(self):
        """Verify list of names cannot be directly modified."""
        with self.assertRaises(AttributeError):
            self.scope.tags.names = 'foo'

    def test_invalid_index(self):
        """Verify accessing a nonexistent tag raises an exception."""
        with self.assertRaises(KeyError):
            self.scope.tags['not_a_tag']


class Tag(object):
    """Base class for testing a tag."""
    def setUp(self):
        e = fixture.parse_xml(self.src_xml)
        self.tag = l5x.tag.Tag(e, None)

        # Initialize a fresh copy of the source XML value, if available.
        try:
            self.src_value = copy.deepcopy(self.xml_value)
        except AttributeError:
            pass

    def test_invalid_desc(self):
        """Ensure non-string types raise an exception."""
        with self.assertRaises(TypeError):
            self.tag.description = 0

    def test_data_type_read(self):
        """Confirm reading the data type returns the correct attribute value."""
        self.assertEqual(self.tag.data_type,
                         self.tag.element.attrib['DataType'])

    def test_data_type_write(self):
        """Confirm an exception is raised when changing the data type."""
        with self.assertRaises(AttributeError):
            self.tag.data_type = 'fail'

    def test_remove_raw_data(self):
        """Ensure setting top-level tag value removes undecorated data."""
        self.tag.value = self.tag.value
        self.assert_no_raw_data_element()

    def assert_no_raw_data_element(self):
        """Confirms any undecorated data element has been removed."""
        data = self.tag.element.findall('Data')
        self.assertEqual(len(data), 1)
        self.assertEqual(data[0].attrib['Format'], 'Decorated')

    def get_value_element(self):
        """Finds the element containing the top-level decorated value.

        The tag name will vary depending on data type, however, it will always
        be the only child of the decorated Data element.
        """
        parent = self.tag.element.find("Data[@Format='Decorated']")
        children = parent.findall('*')
        return children[0]


class Data(unittest.TestCase):
    """Unit tests for the base Data class."""
    class DummyType(l5x.tag.Data):
        """Mock subclass; the Data class is never instantiated directly."""
        pass

    def test_array(self):
        """Confirm array data is delegated to an Array object."""
        element = fixture.parse_xml("""<Array DataType="DINT" Dimensions="1" Radix="Decimal">
<Element Index="[0]" Value="0"/>
</Array>""")
        data = self.DummyType(element, None)
        self.assertIsInstance(data, l5x.tag.Array)

    def test_array_member(self):
        """Confirm array member data is delegated to an ArrayMember object."""
        element = fixture.parse_xml("""<ArrayMember Name="dint_array" DataType="DINT" Dimensions="1" Radix="Decimal">
<Element Index="[0]" Value="0"/>
</ArrayMember>""")
        data = self.DummyType(element, None)
        self.assertIsInstance(data, l5x.tag.ArrayMember)

    def test_name_operand(self):
        """Confirm data identified by Name are separated by dots."""
        e = ElementTree.Element('udt')
        udt = self.DummyType(e, None)

        e = ElementTree.SubElement(udt.element, 'member', {'Name':'foo'})
        member = self.DummyType(e, None, udt)

        e = ElementTree.SubElement(member.element, 'submember', {'Name':'bar'})
        submember = self.DummyType(e, None, member)

        self.assertEqual(submember.operand, '.FOO.BAR')

    def test_index_operand(self):
        """Confirm data identified by Index have no separators."""
        e = ElementTree.Element('array')
        array = self.DummyType(e, None)

        e = ElementTree.SubElement(array.element, 'member', {'Index':'[42]'})
        member = self.DummyType(e, None, array)

        e = ElementTree.SubElement(member.element, 'submember', {'Index':'[0]'})
        submember = self.DummyType(e, None, member)

        self.assertEqual(submember.operand, '[42][0]')


class Integer(Tag):
    """Base class for testing integer data types."""
    def test_length(self):
        """Test len() returns number of bits."""
        self.assertEqual(len(self.tag), self.bits)

    def test_value_read(self):
        """Verify value read returns the attribute converted to an integer."""
        value = 42
        self.set_value(value)
        self.assertEqual(self.tag.value, value)

    def test_value_write(self):
        """Verify new values are properly written to the correct attribute."""
        self.tag.value = 42
        value = self.get_value()
        self.assertEqual(value, 42)

    def test_value_write_type(self):
        """Verify setting the value to a non-integer raises an exception."""
        with self.assertRaises(TypeError):
            self.tag.value = '42'

    def test_value_out_of_range(self):
        """Ensure setting out-of-range values raise an exception."""
        try:
            with self.assertRaises(ValueError):
                self.tag.value = self.value_min - 1
            with self.assertRaises(ValueError):
                self.tag.value = self.value_max + 1

        # Ignore type errors on Python 2.7.x if the out-of-range
        # test value is promoted to long.
        except TypeError:
            pass
                
    def test_value_min(self):
        """Confirm the minimum value is accepted."""
        self.tag.value = self.value_min

    def test_value_max(self):
        """Confirm the maximum value is accepted."""
        self.tag.value = self.value_max

    def test_invalid_bit_indices(self):
        """Verify invalid bit indices raise an exception."""
        with self.assertRaises(IndexError):
            self.tag[-1]
        with self.assertRaises(IndexError):
            self.tag[self.bits]

    def test_bit_index_type(self):
        """Verify non-integer bit indices raise an exception."""
        with self.assertRaises(TypeError):
            self.tag['foo']

    def test_bit_value_read(self):
        """Confirm non-sign bits reflect the current integer value."""
        for bit in range(self.bits - 1):
            # Check setting only the target bit.
            value = 1 << bit
            self.set_value(value)
            for test_bit in range(self.bits):
                if test_bit == bit:
                    test_value = 1
                else:
                    test_value = 0
                self.assertEqual(self.tag[test_bit].value, test_value)

            # Check setting all except the target bit.
            value = ~(1 << bit)
            self.set_value(value)
            for test_bit in range(self.bits):
                if test_bit == bit:
                    test_value = 0
                else:
                    test_value = 1
                self.assertEqual(self.tag[test_bit].value, test_value)

    def test_bit_value_write(self):
        """Confirm writing non-sign bits properly update the integer value."""
        for bit in range(self.bits - 1):
            # Check setting only the target bit.
            self.set_value(0)
            self.tag[bit].value = 1
            value = self.get_value()
            self.assertEqual(value, 1 << bit)

            # Check clearing only the target bit.
            self.set_value(-1)
            self.tag[bit].value = 0
            value = self.get_value()
            self.assertEqual(value, ~(1 << bit))

    def test_sign_bit_read(self):
        """Confirm MSB is treated as the sign when reading a value."""
        sign_bit = self.bits - 1
        min_negative = -(1 << sign_bit)
        self.set_value(min_negative)
        self.assertEqual(self.tag[sign_bit].value, 1)

    def test_sign_bit_write(self):
        """Confirm MSB is treated as the sign bit when writing a value."""
        sign_bit = self.bits - 1
        min_negative = -(1 << sign_bit)
        self.tag[sign_bit].value = 1
        value = self.get_value()
        self.assertEqual(value, min_negative)

    def test_bit_write_type(self):
        """Confirm an exception is raised when setting a bit to a non-integer."""
        with self.assertRaises(TypeError):
            self.tag[0].value = 'foo'

    def test_bit_value_range(self):
        """Ensure bit values other than 0 or 1 raise an exception."""
        with self.assertRaises(ValueError):
            self.tag[0].value = -1
        with self.assertRaises(ValueError):
            self.tag[0].value = 2

    def test_bit_desc_read(self):
        """Confirm reading an existing bit description."""
        for bit in range(self.bits):
            comment_text = "comment for bit {0}".format(bit)
            self.add_bit_description(bit, comment_text)
            self.assertEqual(self.tag[bit].description, comment_text)

    def test_bit_desc_read_none(self):
        """Confirm reading a nonexistent bit description."""
        for bit in range(self.bits):
            self.assertIsNone(self.tag[bit].description)

    def test_bit_desc_read_none_other(self):
        """Confirm reading a nonexistent bit comment when other comments exist."""
        for bit in range(self.bits):
            self.assertIsNone(self.tag[bit].description)

            # Add the comment after checking to create unrelated comments
            # for the next bit.
            comment_text = "comment for bit {0}".format(bit)
            self.add_bit_description(bit, comment_text)

    def test_bit_desc_write_new(self):
        """Confirm creating a new bit description."""
        for bit in range(self.bits):
            comment = "bit {0}".format(bit)
            self.tag[bit].description = comment
            self.assert_bit_description(bit, comment)

    def test_bit_desc_overwrite(self):
        """Confirm overwriting an existing bit description."""
        for bit in range(self.bits):
            old = "old bit {0} comment".format(bit)
            self.add_bit_description(bit, old)
            new = "new bit {0} comment".format(bit)
            self.tag[bit].description = new
            self.assert_bit_description(bit, new)

    def test_bit_desc_del(self):
        """Confirm removal of an existing bit description."""
        for bit in range(self.bits):
            desc = "bit {0} comment".format(bit)
            self.add_bit_description(bit, desc)
            self.tag[bit].description = None
            path = "Comments/Comment[@Operand='.{0}']".format(bit)
            comment = self.tag.element.findall(path)
            self.assertEqual(len(comment), 0)

    def test_bit_value_raw_data(self):
        """Ensure undecorated data is cleared when setting a single bit."""
        self.tag[0].value = 1
        self.assert_no_raw_data_element()

    def get_value(self):
        """Returns the value currently stored in the XML attribute."""
        element = self.get_value_element()
        return int(element.attrib['Value'])

    def set_value(self, value):
        """Updates the value stored in the XML attribute."""
        element = self.get_value_element()
        element.attrib['Value'] = str(value)

    def add_bit_description(self, bit, text):
        """Creates a bit description."""
        comments = self.tag.element.find('Comments')
        if comments is None:
            comments = ElementTree.SubElement(self.tag.element, 'Comments')
        operand = ".{0}".format(bit)
        comment = ElementTree.SubElement(comments, 'Comment',
                                         {'Operand':operand})
        cdata = ElementTree.SubElement(comment, dom.CDATA_TAG)
        cdata.text = text

    def assert_bit_description(self, bit, text):
        """Tests to ensure a description for a specific bit exists."""
        # Confirm a single Comments element under the tag parent.
        comments = self.tag.element.findall('Comments')
        self.assertEqual(len(comments), 1)

        # Confirm a single Comment with matching operand.
        path = "Comment[@Operand='.{0}']".format(bit)
        comment = comments[0].findall(path)
        self.assertEqual(len(comment), 1)

        # Confirm a single CDATA child with matching text.
        cdata = [e for e in comment[0].iter()]
        del cdata[0] # Exclude the Comment parent.
        self.assertEqual(len(cdata), 1)
        self.assertEqual(cdata[0].tag, dom.CDATA_TAG)
        self.assertEqual(cdata[0].text, text)


class TestSINT(Integer, unittest.TestCase):
    src_xml = """<Tag Name="sint" TagType="Base" DataType="SINT" Radix="Decimal" Constant="false" ExternalAccess="Read/Write">
<Data>00</Data>
<Data Format="Decorated">
<DataValue DataType="SINT" Radix="Decimal" Value="0"/>
</Data>
</Tag>"""

    xml_value = 0
    bits = 8
    value_min = -128
    value_max = 127


class TestINT(Integer, unittest.TestCase):
    src_xml = """<Tag Name="int" TagType="Base" DataType="INT" Radix="Decimal" Constant="false" ExternalAccess="Read/Write">
<Data>00 00</Data>
<Data Format="Decorated">
<DataValue DataType="INT" Radix="Decimal" Value="0"/>
</Data>
</Tag>"""

    xml_value = 0
    bits = 16
    value_min = -32768
    value_max = 32767


class TestDINT(Integer, unittest.TestCase):
    src_xml = """<Tag Name="dint" TagType="Base" DataType="DINT" Radix="Decimal" Constant="false" ExternalAccess="Read/Write">
<Data>00 00 00 00</Data>
<Data Format="Decorated">
<DataValue DataType="DINT" Radix="Decimal" Value="0"/>
</Data>
</Tag>"""

    xml_value = 0
    bits = 32
    value_min = -2147483648
    value_max = 2147483647


class TestBOOL(Tag, unittest.TestCase):
    """BOOL type tests."""
    src_xml = """<Tag Name="bool" TagType="Base" DataType="BOOL" Radix="Decimal" Constant="false" ExternalAccess="Read/Write">
<Data>00</Data>
<Data Format="Decorated">
<DataValue DataType="BOOL" Radix="Decimal" Value="0"/>
</Data>
</Tag>"""

    xml_value = 0

    def test_value_read(self):
        """Confirm reading the current value."""
        value = self.get_value_element()
        value.attrib['Value'] = '1'
        self.assertEqual(self.tag.value, 1)

    def test_value_write(self):
        """Confirm writing a new value."""
        self.tag.value = 1
        value = self.get_value_element()
        self.assertEqual(int(value.attrib['Value']), 1)

    def test_value_invalid_type(self):
        """Confirm an exception is raised when writing a non-integer value."""
        with self.assertRaises(TypeError):
            self.tag.value = False

    def test_out_of_range_value(self):
        """Test exception when setting values other than 0 or 1."""
        for x in [-1, 2]:
            with self.assertRaises(ValueError):
                self.tag.value = x


class TestREAL(Tag, unittest.TestCase):
    """REAL type tests."""
    src_xml = """<Tag Name="real" TagType="Base" DataType="REAL" Radix="Float" Constant="false" ExternalAccess="Read/Write">
<Data>00 00 00 00</Data>
<Data Format="Decorated">
<DataValue DataType="REAL" Radix="Float" Value="0.0"/>
</Data>
</Tag>"""

    xml_value = 0.0

    def test_value_read(self):
        """Confirm reading the current value."""
        value = self.get_value_element()
        value.attrib['Value'] = str(math.pi)
        self.assertAlmostEqual(self.tag.value, math.pi)

    def test_value_write(self):
        """Confirm writing a new value."""
        self.tag.value = math.e
        value = self.get_value_element()
        self.assertAlmostEqual(float(value.attrib['Value']), math.e)

    def test_value_invalid_type(self):
        """Confirm an exception is raised when writing a non-float value."""
        with self.assertRaises(TypeError):
            self.tag.value = 'not a float'

    def test_invalid_values(self):
        """Ensure NaN and infinite values raise an exception."""
        for value in ['NaN', 'inf']:
            with self.assertRaises(ValueError):
                self.tag.value = float(value)


class TestSingleDimensionalArray(Tag, unittest.TestCase):
    """Single-dimensional array tests."""
    src_xml = """<Tag Name="array" TagType="Base" DataType="DINT" Dimensions="3" Radix="Decimal" Constant="false" ExternalAccess="Read/Write">
<Data>00 00 00 00 00 00 00 00 00 00 00 00</Data>
<Data Format="Decorated">
<Array DataType="DINT" Dimensions="3" Radix="Decimal">
<Element Index="[0]" Value="0"/>
<Element Index="[1]" Value="0"/>
<Element Index="[2]" Value="0"/>
</Array>
</Data>
</Tag>"""

    xml_value = [0, 0, 0]

    def test_shape(self):
        """Ensure shape is a tuple with the correct dimensions.."""
        self.assertEqual(self.tag.shape, (3,))
            
    def test_index_type(self):
        """Ensure non-integer indices raise an exception."""
        with self.assertRaises(TypeError):
            self.tag['not an int']
            
    def test_index_range(self):
        """Ensure negative and indices beyond the end raise exceptions."""
        for i in [-1, self.tag.shape[0]]:
            with self.assertRaises(IndexError):
                self.tag[i]

    def test_value_read(self):
        """Confirm reading the top-level value returns a list of values."""
        new = [100 + i for i in range(self.tag.shape[0])]
        for i in range(len(new)):
            element = self.get_value_element(i)
            element.attrib['Value'] = str(new[i])
        self.assertEqual(self.tag.value, new)

    def test_element_value_read(self):
        """Confirm reading a single value."""
        for i in range(self.tag.shape[0]):
            value = i + 10
            element = self.get_value_element(i)
            element.attrib['Value'] = str(value)
            self.assertEqual(self.tag[i].value, value)

    def test_value_write(self):
        """Confirm setting a new value to all elements with a list."""
        new = [100 + i for i in range(self.tag.shape[0])]
        self.tag.value = new
        for i in range(len(new)):
            element = self.get_value_element(i)
            value = int(element.attrib['Value'])
            self.assertEqual(value, new[i])

    def test_value_write_short(self):
        """Confirm setting a value to a list with fewer elements starts overwriting at the beginning."""
        new = [100 + i for i in range(self.tag.shape[0] - 1)]
        self.tag.value = new
        for i in range(self.tag.shape[0]):
            element = self.get_value_element(i)
            try:
                value = new[i]
            except IndexError:
                value = 0
            self.assertEqual(value, int(element.attrib['Value']))

    def test_value_write_too_long(self):
        """Confirm an exception is raised when setting the value to a list that is too long."""
        with self.assertRaises(IndexError):
            self.tag.value = [0] * (self.tag.shape[0] + 1)

    def test_element_value_write(self):
        """Confirm writing a single element."""
        for i in range(self.tag.shape[0]):
            new_value = i * 2
            self.tag[i].value = new_value
            element = self.get_value_element(i)
            self.assertEqual(new_value, int(element.attrib['Value']))

    def test_invalid_value_type(self):
        """Test setting value to a non-list raises an exception."""
        with self.assertRaises(TypeError):
            self.tag.value = 'not a list'

    def test_element_description_read(self):
        """Test reading an existing element description."""
        desc = 'some comment'
        self.set_comment(0, desc)
        self.assertEqual(self.tag[0].description, desc)

    def test_element_description_read_none(self):
        """Test reading a nonexistent element description."""
        self.assertIsNone(self.tag[0].description)

    def test_element_description_new(self):
        """Test creating a new element description."""
        self.tag[0].description = 'foo'
        desc = self.get_comment(0)
        self.assertEqual(desc, 'foo')

    def test_element_description_overwrite(self):
        """Test overwriting an existing element description."""
        self.set_comment(0, 'old')
        self.tag[0].description = 'new'
        desc = self.get_comment(0)
        self.assertEqual(desc, 'new')

    def test_element_description_delete(self):
        """Test removing an existing element description."""
        self.set_comment(0, 'spam')
        self.tag[0].description = None
        desc = self.get_comment(0)
        self.assertIsNone(desc)

    def test_element_description_delete_none(self):
        """Test removing a nonexistent element description."""
        self.tag[0].description = None
        desc = self.get_comment(0)
        self.assertIsNone(desc)

    def test_element_value_raw_data(self):
        """Ensure setting a single element clears undecorated data."""
        self.tag[0].value = 0
        self.assert_no_raw_data_element()

    def get_value_element(self, index):
        """Finds the element containing an element value."""
        path = "Data/Array/Element[@Index='[{0}]']".format(index)
        return self.tag.element.find(path)

    def set_comment(self, index, text):
        """Creates a comment for a single element."""
        comments = ElementTree.SubElement(self.tag.element, 'Comments')
        operand = "[{0}]".format(index)
        comment = ElementTree.SubElement(comments, 'Comment',
                                         {'Operand':operand})
        cdata = ElementTree.SubElement(comment, dom.CDATA_TAG)
        cdata.text = text

    def get_comment(self, index):
        """Finds a comment for a single element."""
        path = "Comments/Comment[@Operand='[{0}]']/{1}".format(index,
                                                               dom.CDATA_TAG)
        element = self.tag.element.find(path)
        try:
            return element.text
        except AttributeError:
            return None


class TestMultiDimensionalArray(Tag, unittest.TestCase):
    """Multi-dimensional array tests"""
    src_xml = """<Tag Name="array" TagType="Base" DataType="DINT" Dimensions="2 3 4" Radix="Decimal" Constant="false" ExternalAccess="Read/Write">
<Data>01 00 00 00 02 00 00 00 03 00 00 00 04 00 00 00 
05 00 00 00 06 00 00 00 07 00 00 00 08 00 00 00 
09 00 00 00 0A 00 00 00 0B 00 00 00 0C 00 00 00 
0D 00 00 00 0E 00 00 00 0F 00 00 00 10 00 00 00 
11 00 00 00 12 00 00 00 13 00 00 00 14 00 00 00 
15 00 00 00 16 00 00 00 17 00 00 00 18 00 00 00</Data>
<Data Format="Decorated">
<Array DataType="DINT" Dimensions="2,3,4" Radix="Decimal">
<Element Index="[0,0,0]" Value="1"/>
<Element Index="[0,0,1]" Value="2"/>
<Element Index="[0,0,2]" Value="3"/>
<Element Index="[0,0,3]" Value="4"/>
<Element Index="[0,1,0]" Value="5"/>
<Element Index="[0,1,1]" Value="6"/>
<Element Index="[0,1,2]" Value="7"/>
<Element Index="[0,1,3]" Value="8"/>
<Element Index="[0,2,0]" Value="9"/>
<Element Index="[0,2,1]" Value="10"/>
<Element Index="[0,2,2]" Value="11"/>
<Element Index="[0,2,3]" Value="12"/>
<Element Index="[1,0,0]" Value="13"/>
<Element Index="[1,0,1]" Value="14"/>
<Element Index="[1,0,2]" Value="15"/>
<Element Index="[1,0,3]" Value="16"/>
<Element Index="[1,1,0]" Value="17"/>
<Element Index="[1,1,1]" Value="18"/>
<Element Index="[1,1,2]" Value="19"/>
<Element Index="[1,1,3]" Value="20"/>
<Element Index="[1,2,0]" Value="21"/>
<Element Index="[1,2,1]" Value="22"/>
<Element Index="[1,2,2]" Value="23"/>
<Element Index="[1,2,3]" Value="24"/>
</Array>
</Data>
</Tag>"""

    xml_value = [
        [
            [1, 2, 3, 4],
            [5, 6, 7, 8],
            [9, 10, 11, 12]
        ],
        [
            [13, 14, 15, 16],
            [17, 18, 19, 20],
            [21, 22, 23, 24]
        ]
    ]

    def test_shape_values(self):
        """Verify correct dimension values."""
        self.assertEqual(self.tag.shape[0], len(self.src_value[0][0]))
        self.assertEqual(self.tag.shape[1], len(self.src_value[0]))
        self.assertEqual(self.tag.shape[2], len(self.src_value))

    def test_value_read(self):
        """Verify reading values for each dimension."""
        self.assertEqual(self.tag.value, self.src_value)

        for i in range(len(self.tag.value)):
            self.assertEqual(self.tag.value[i], self.src_value[i])

            for j in range(len(self.tag.value[i])):
                self.assertEqual(self.tag.value[i][j], self.src_value[i][j])

    def test_value_write_single(self):
        """Verify writing a single element value."""
        self.src_value[1][2][3] = 123
        self.tag[1][2][3].value = 123
        self.assert_element_values()

    def test_value_write_subarray(self):
        """Verify writing a subarray value."""
        self.src_value[0][0] = [1000, 1001, 1002, 1003]
        self.tag[0][0].value = [1000, 1001, 1002, 1003]
        self.assert_element_values()

    def test_value_write_all(self):
        """Verify writing a nested list to the top-level value."""
        new_values = [
            [
                [-1, -2, -3, -4],
                [-5, -6, -7, -8],
                [-9, -10, -11, -12]
            ],
            [
                [-13, -14, -15, -16],
                [-17, -18, -19, -20],
                [-21, -22, -23, -24]
            ]
        ]
        self.src_value = new_values
        self.tag.value = new_values
        self.assert_element_values()

    def assert_element_values(self):
        """Confirms all element values match the source array."""
        # Iterate though all the source array values and confirm a
        # matching XML element value.
        indices = [range(self.tag.shape[x]) for x in range(len(self.tag.shape))]
        indices.reverse()
        for subscript in itertools.product(*indices):
            # Acquire the value stored in the XML attribute.
            index = "[{0}]".format(','.join([str(i) for i in subscript]))
            element = self.tag.element.find(
                'Data/Array/*[@Index="{0}"]'.format(index))
            xml_value = int(element.attrib['Value'])

            # Descend through the source array to select the single value
            # for the given subscript.
            src_value = self.src_value
            for i in subscript:
                src_value = src_value[i]

            self.assertEqual(xml_value, src_value)

        # Iterate through all the XML value elements and confirm a
        # matching source array value.
        for element in self.tag.element.findall('Data/Array/*'):
            index = [int(x) for x in element.attrib['Index'][1:-1].split(',')]
            xml_value = int(element.attrib['Value'])

            # Descend through the source array to select the single value
            # for the given index.
            src_value = self.src_value
            for i in index:
                src_value = src_value[i]

            self.assertEqual(xml_value, src_value)

    def test_subarray_description(self):
        """Confirm descriptions are not permitted for subarrays."""
        ar = self.tag
        for i in range(len(self.tag.shape) - 1):
            ar = ar[0]
            with self.assertRaises(TypeError):
                ar.description
            with self.assertRaises(TypeError):
                ar.description = 'test'


class ArrayResize(Tag):
    """Base class for array resizing tests."""
    def test_shape(self):
        """Ensure the tag's shape value is updated."""
        self.resize()
        self.assertEqual(self.tag.shape, self.dim)

    def test_tag_dimensions_attr(self):
        """Ensure the top-level Tag element's Dimensions attribute is set."""
        self.resize()
        dims = ' '.join(reversed([str(x) for x in self.dim]))
        self.assertEqual(dims, self.tag.element.attrib['Dimensions'])

    def test_array_dimensions_attr(self):
        """Ensures the Array element's Dimensions attribute is set."""
        self.resize()
        dims = ','.join(reversed([str(x) for x in self.dim]))
        array = self.tag.element.find('Data/Array')
        self.assertEqual(dims, array.attrib['Dimensions'])

    def test_raw_data_removed(self):
        """Ensure the original raw data array is deleted."""
        self.resize()
        self.assert_no_raw_data_element()

    def test_indices(self):
        """Confirm element indices match the new shape."""
        self.resize()

        # Get indices from all the XML elements.
        array = self.tag.element.find('Data/Array')
        xml_idx = set([e.attrib['Index'] for e in array])

        # Create indices from the expected dimensions.
        ranges = [range(self.dim[i]) for i in range(len(self.dim))]
        ranges.reverse()
        new_idx = set()
        for idx in itertools.product(*ranges):
            new_idx.add("[{0}]".format(','.join([str(x) for x in idx])))

        self.assertEqual(xml_idx, new_idx)


class ArrayResizeAddDimension(ArrayResize, unittest.TestCase):
    """Tests for resizing an array by adding a dimension."""
    src_xml = """<Tag Name="array" TagType="Base" DataType="DINT" Dimensions="2" Radix="Decimal" Constant="false" ExternalAccess="Read/Write">
<Data>00 00 00 00 01 00 00 00</Data>
<Data Format="Decorated">
<Array DataType="DINT" Dimensions="2" Radix="Decimal">
<Element Index="[0]" Value="0"/>
<Element Index="[1]" Value="1"/>
</Array>
</Data>
</Tag>"""

    xml_value = [0, 1]

    def resize(self):
        """Adds a new dimension."""
        self.dim = (3, 4)
        self.tag.shape = self.dim


class ArrayResizeRemoveDimension(ArrayResize, unittest.TestCase):
    """Tests for resizing an array by removing a dimension."""
    src_xml = """<Tag Name="array" TagType="Base" DataType="DINT" Dimensions="3 2" Radix="Decimal" Constant="false" ExternalAccess="Read/Write">
<Data>00 00 00 00 01 00 00 00 02 00 00 00 03 00 00 00 
04 00 00 00 05 00 00 00</Data>
<Data Format="Decorated">
<Array DataType="DINT" Dimensions="3,2" Radix="Decimal">
<Element Index="[0,0]" Value="0"/>
<Element Index="[0,1]" Value="1"/>
<Element Index="[1,0]" Value="2"/>
<Element Index="[1,1]" Value="3"/>
<Element Index="[2,0]" Value="4"/>
<Element Index="[2,1]" Value="5"/>
</Array>
</Data>
</Tag>"""

    xml_value = [
        [0, 1],
        [2, 3],
        [4, 5]
    ]

    def resize(self):
        """Removes a dimension."""
        self.dim = (2,)
        self.tag.shape = self.dim


class ArrayResizeEnlarge(ArrayResize, unittest.TestCase):
    """Tests for resizing an array by enlarging a dimension."""
    src_xml = """<Tag Name="array" TagType="Base" DataType="DINT" Dimensions="3 2" Radix="Decimal" Constant="false" ExternalAccess="Read/Write">
<Data>00 00 00 00 01 00 00 00 02 00 00 00 03 00 00 00 
04 00 00 00 05 00 00 00</Data>
<Data Format="Decorated">
<Array DataType="DINT" Dimensions="3,2" Radix="Decimal">
<Element Index="[0,0]" Value="0"/>
<Element Index="[0,1]" Value="1"/>
<Element Index="[1,0]" Value="2"/>
<Element Index="[1,1]" Value="3"/>
<Element Index="[2,0]" Value="4"/>
<Element Index="[2,1]" Value="5"/>
</Array>
</Data>
</Tag>"""

    xml_value = [
        [0, 1],
        [2, 3],
        [4, 5]
    ]

    def resize(self):
        """Enlarges a dimension."""
        self.dim = (2, 4)
        self.tag.shape = self.dim


class ArrayResizeReduce(ArrayResize, unittest.TestCase):
    """Tests for resizing an array by shrinking a dimension."""
    src_xml = """<Tag Name="array" TagType="Base" DataType="DINT" Dimensions="4 2" Radix="Decimal" Constant="false" ExternalAccess="Read/Write">
<Data>00 00 00 00 01 00 00 00 02 00 00 00 03 00 00 00 
04 00 00 00 05 00 00 00 06 00 00 00 07 00 00 00</Data>
<Data Format="Decorated">
<Array DataType="DINT" Dimensions="4,2" Radix="Decimal">
<Element Index="[0,0]" Value="0"/>
<Element Index="[0,1]" Value="1"/>
<Element Index="[1,0]" Value="2"/>
<Element Index="[1,1]" Value="3"/>
<Element Index="[2,0]" Value="4"/>
<Element Index="[2,1]" Value="5"/>
<Element Index="[3,0]" Value="6"/>
<Element Index="[3,1]" Value="7"/>
</Array>
</Data>
</Tag>"""

    xml_value = [
        [0, 1],
        [2, 3],
        [4, 5],
        [6, 7]
    ]

    def resize(self):
        """Shrinks a dimension."""
        self.dim = (2, 3)
        self.tag.shape = self.dim


class ArrayResizeInvalid(Tag, unittest.TestCase):
    """Tests for illegal array resizing."""
    src_xml = """<Tag Name="array" TagType="Base" DataType="DINT" Dimensions="3" Radix="Decimal" Constant="false" ExternalAccess="Read/Write">
<Data>01 00 00 00 02 00 00 00 03 00 00 00</Data>
<Data Format="Decorated">
<Array DataType="DINT" Dimensions="3" Radix="Decimal">
<Element Index="[0]" Value="1"/>
<Element Index="[1]" Value="2"/>
<Element Index="[2]" Value="3"/>
</Array>
</Data>
</Tag>"""

    xml_value = [1, 2, 3]

    def test_too_many_dimensions(self):
        """Confirm an exception is raised for more than 3 dimensions."""
        with self.assertRaises(ValueError):
            self.tag.shape = (1, 2, 3, 4)

    def test_zero_dimensions(self):
        """Confirm an exception is raised for no dimensions."""
        with self.assertRaises(ValueError):
            self.tag.shape = ()

    def test_non_tuple(self):
        """Confirm an exception is raised for a non-tuple shape."""
        with self.assertRaises(TypeError):
            self.tag.shape = 42

    def test_negative_size(self):
        """Confirm an exception is raised for a negative dimension."""
        with self.assertRaises(ValueError):
            self.tag.shape = (-1,)

    def test_zero_size(self):
        """Confirm an exception is raised for a dimension of zero."""
        with self.assertRaises(ValueError):
            self.tag.shape = (0,)

    def test_non_integer(self):
        """Confirm an exception is raised for a non-integer dimension."""
        with self.assertRaises(TypeError):
            self.tag.shape = (5.0,)


class Structure(Tag, unittest.TestCase):
    """Structured data tag tests."""
    src_xml = """<Tag Name="timer" TagType="Base" DataType="TIMER" Constant="false" ExternalAccess="Read/Write">
<Data>00 00 00 E0 FF FF FF FF FE FF FF FF</Data>
<Data Format="Decorated">
<Structure DataType="TIMER">
<DataValueMember Name="PRE" DataType="DINT" Radix="Decimal" Value="-1"/>
<DataValueMember Name="ACC" DataType="DINT" Radix="Decimal" Value="-2"/>
<DataValueMember Name="EN" DataType="BOOL" Value="1"/>
<DataValueMember Name="TT" DataType="BOOL" Value="1"/>
<DataValueMember Name="DN" DataType="BOOL" Value="1"/>
</Structure>
</Data>
</Tag>"""

    xml_value = {
        'PRE':-1,
        'ACC':-2,
        'EN':1,
        'TT':1,
        'DN':1
    }

    def test_invalid_value_type(self):
        """Test setting value to a non-dict raises an exception."""
        with self.assertRaises(TypeError):
            self.tag.value = 'not a dict'

    def test_value_read(self):
        """Confirm reading the top-level value returns a correct dict."""
        self.assertDictEqual(self.tag.value, self.src_value)

    def test_value_write(self):
        """Confirm writing a dict to the top-level value.."""
        self.src_value = {'PRE':42, 'ACC':142, 'EN':0, 'TT':0, 'DN':0}
        self.tag.value = self.src_value
        self.assert_member_values()

    def test_value_write_partial(self):
        """Confirm writing a partial dict to the top-level value."""
        new_values = {'PRE':-100}
        self.tag.value = new_values
        self.src_value.update(new_values)
        self.assert_member_values()

    def test_nonstring_index(self):
        """Verify non-string indices raise an exception."""
        with self.assertRaises(TypeError):
            self.tag[0].value

    def test_invalid_index(self):
        """Verify indices for nonexistent members raise an exception."""
        with self.assertRaises(KeyError):
            self.tag['foo'].value

    def test_member_value_read(self):
        """Test reading individual member values."""
        for name in self.src_value:
            self.assertEqual(self.tag[name].value, self.src_value[name])

    def test_member_value_write(self):
        """Test writing individual member values."""
        new_values = {
            'PRE':200,
            'ACC':300,
            'EN':0,
            'TT':0,
            'DN':0
        }

        for name in self.src_value:
            self.tag[name].value = new_values[name]
            self.src_value[name] = new_values[name]
            self.assert_member_values()

    def test_names_read(self):
        """Confirm the names attributes contains all members."""
        src_names = set(self.src_value.keys())
        tag_names = set(self.tag.names)
        self.assertEqual(src_names, tag_names)

    def test_names_write(self):
        """
        Confirm an exception is raised when attempting to write to the
        names attribute.
        """
        with self.assertRaises(AttributeError):
            self.tag.names = 'fail'

    def test_write_clear_raw_data(self):
        """
        Confirm writing a dict to the top-level value clears undecorated data.
        """
        self.tag.value = self.src_value
        self.assert_no_raw_data_element()

    def test_member_write_clear_raw_data(self):
        """Ensure setting a single member clears undecorated data."""
        self.tag['PRE'].value = 0
        self.assert_no_raw_data_element()

    def assert_member_values(self):
        """Confirms the value of all data members match the source dict."""
        st = self.tag.element.find('Data/Structure')

        # Confirm all XML values match the source dict.
        for e in st:
            value = int(e.attrib['Value'])
            name = e.attrib['Name']
            self.assertEqual(self.src_value[name], value)

        # Confirm all source dict members match the XML values.
        for name in self.src_value:
            e = st.find('*[@Name="{0}"]'.format(name))
            value = int(e.attrib['Value'])
            self.assertEqual(self.src_value[name], value)


class Compound(Tag, unittest.TestCase):
    """Tests for a data types containing nested arrays and structures."""
    src_xml = """<Tag Name="udt" TagType="Base" DataType="udt" Constant="false" ExternalAccess="Read/Write">
<Data>00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 
00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 
00 00 00 00 00 00 00 00</Data>
<Data Format="Decorated">
<Structure DataType="udt">
<ArrayMember Name="dint_array" DataType="DINT" Dimensions="10" Radix="Decimal">
<Element Index="[0]" Value="0"/>
<Element Index="[1]" Value="0"/>
<Element Index="[2]" Value="0"/>
<Element Index="[3]" Value="0"/>
<Element Index="[4]" Value="0"/>
<Element Index="[5]" Value="0"/>
<Element Index="[6]" Value="0"/>
<Element Index="[7]" Value="0"/>
<Element Index="[8]" Value="0"/>
<Element Index="[9]" Value="0"/>
</ArrayMember>
</Structure>
</Data>
</Tag>"""

    def test_member_array_resize(self):
        """Ensure member arrays cannot be resized."""
        with self.assertRaises(AttributeError):
            self.tag['dint_array'].shape = (1,)


class Base(unittest.TestCase):
    """Tests for base, i.e. not produced or consumed, tags."""
    def setUp(self):
        """Creates a mock base tag."""
        e = fixture.parse_xml("""<Tag Name="dint" TagType="Base" DataType="DINT" Radix="Decimal" Constant="false" ExternalAccess="Read/Write">
<Data>00 00 00 00</Data>
<Data Format="Decorated">
<DataValue DataType="DINT" Radix="Decimal" Value="0"/>
</Data>
</Tag>""")
        self.tag = l5x.tag.Tag(e, None)

    def test_producer_read(self):
        """Confirm reading the producer attribute raises an exception."""
        with self.assertRaises(TypeError):
            self.tag.producer

    def test_producer_write(self):
        """Confirm writing the producer attribute raises an exception."""
        with self.assertRaises(TypeError):
            self.tag.producer = 'foo'

    def test_remote_tag_read(self):
        """Confirm reading the remote tag attribute raises an exception."""
        with self.assertRaises(TypeError):
            self.tag.remote_tag

    def test_remote_tag_write(self):
        """Confirm writing the remote tag attribute raises an exception."""
        with self.assertRaises(TypeError):
            self.tag.remote_tag = 'foo'

    def test_description_element_order(self):
        """Ensure a description is created as tag element's first child."""
        self.tag.description = 'description'
        first_child = self.tag.element.find('*')
        self.assertEqual(first_child.tag, 'Description')


class Consumed(unittest.TestCase):
    """Tests for attributes specific to consumed tags."""
    def setUp(self):
        e = fixture.parse_xml("""<Tag Name="dint" TagType="Consumed" DataType="DINT" Radix="Decimal" ExternalAccess="Read/Write">
<ConsumeInfo Producer="producer" RemoteTag="source_tag" RemoteInstance="0" RPI="20"/>
<Data>00 00 00 00</Data>
<ForceData>00 00 00 00 00 00 00 00 00 00 00 00</ForceData>
<Data Format="Decorated">
<DataValue DataType="DINT" Radix="Decimal" Value="0"/>
</Data>
</Tag>""")
        self.tag = l5x.tag.Tag(e, None)

    def test_get_producer(self):
        """Confirm producer returns the correct attribute value."""
        self.assertEqual(self.tag.producer, 'producer')

    def test_set_producer(self):
        """Confirm setting the producer alters the correct attribute."""
        self.tag.producer = 'spam'
        consume_info = self.tag.element.find('ConsumeInfo')
        self.assertEqual(consume_info.attrib['Producer'], 'spam')

    def test_get_remote_tag(self):
        """Confirm remote tag returns the correct attribute value."""
        self.assertEqual(self.tag.remote_tag, 'source_tag')

    def test_set_remote_tag(self):
        """Confirm setting the remote tag alters the correct attribute."""
        self.tag.remote_tag = 'eggs'
        consume_info = self.tag.element.find('ConsumeInfo')
        self.assertEqual(consume_info.attrib['RemoteTag'], 'eggs')

    def test_set_nonstring(self):
        """Confirm an exception is raised when setting to a non-string."""
        with self.assertRaises(TypeError):
            self.tag.producer = 0

    def test_set_empty(self):
        """Confirm an exception is raised when setting to an empty string."""
        with self.assertRaises(ValueError):
            self.tag.remote_tag = ''

    def test_not_consumed(self):
        """Confirm access to a base tag raises an exception."""
        self.tag.element.attrib['TagType'] = 'Base'
        with self.assertRaises(TypeError):
            self.tag.producer

    def test_element_order(self):
        """Ensure a Description element is placed after ConsumedInfo."""
        self.tag.description = 'description'
        child_tags = [e.tag for e in self.tag.element.findall('*')]
        self.assertLess(child_tags.index('ConsumeInfo'),
                        child_tags.index('Description'))


class LanguageBase(unittest.TestCase):
    """Base class for tests involving multilanguage comments and descriptions."""
    TARGET_LANGUAGE = 'en-US'

    def create_tag(self, xml_str):
        e = fixture.parse_xml(xml_str)
        self.tag = l5x.tag.Tag(e, None)

    def set_multilanguage(self):
        """Enables multilingual comments."""
        self.tag.lang = self.TARGET_LANGUAGE

    def assert_cdata_content(self, parent, text):
        """
        Verifies a given element contains a single CDATA subelement with a
        matching string. Note, this tests for a CDATA element, not an
        actual CDATA section, as unit tests using this function operate
        using CDATA elements. See CDATA_TAG comments in the dom module
        for details.
        """
        # Confirm the parent element contains a single CDATA subelement.
        children = [e for e in parent.iterfind('*')]
        self.assertEqual(len(children), 1)
        self.assertEqual(children[0].tag, dom.CDATA_TAG)

        # Confirm the content of the new CDATA section.
        self.assertEqual(children[0].text, text)

    def assert_no_matching_element(self, path):
        """Verifies a given XPath does not match under the mock tag element."""
        result = self.tag.element.findall(path)
        self.assertEqual(len(result), 0)


class DescriptionLanguage(LanguageBase):
    """Tests for multilanguage descriptions."""
    def test_single_read(self):
        """Confirm reading a description from a single-language project."""
        self.create_tag("""<Tag Name="tag" TagType="Base" DataType="DINT" Radix="Decimal" Constant="false" ExternalAccess="Read/Write">
<Description>
<![CDATA[foo]]>
</Description>
<Data>00 00 00 00</Data>
<Data Format="Decorated">
<DataValue DataType="DINT" Radix="Decimal" Value="0"/>
</Data>
</Tag>""")
        self.assertEqual(self.tag.description, 'foo')

    def test_multi_read(self):
        """
        Confirm reading a description from a multi-language project returns
        only content from the current language.
        """
        self.create_tag("""<Tag Name="tag" TagType="Base" DataType="DINT" Radix="Decimal" Constant="false" ExternalAccess="Read/Write">
<Description>
<LocalizedDescription Lang="en-US">
<![CDATA[pass]]>
</LocalizedDescription>
<LocalizedDescription Lang="zh-CN">
<![CDATA[fail]]>
</LocalizedDescription>
</Description>
<Data>00 00 00 00</Data>
<Data Format="Decorated">
<DataValue DataType="DINT" Radix="Decimal" Value="0"/>
</Data>
</Tag>""")
        self.set_multilanguage()
        self.assertEqual(self.tag.description, 'pass')

    def test_single_read_none(self):
        """
        Confirm reading an empty description from a single-language project.
        """
        self.create_tag("""<Tag Name="tag" TagType="Base" DataType="DINT" Radix="Decimal" Constant="false" ExternalAccess="Read/Write">
<Data>00 00 00 00</Data>
<Data Format="Decorated">
<DataValue DataType="DINT" Radix="Decimal" Value="0"/>
</Data>
</Tag>""")
        self.assertIsNone(self.tag.description)

    def test_multi_read_none(self):
        """
        Confirm reading an empty description from a multi-language project.
        """
        self.create_tag("""<Tag Name="tag" TagType="Base" DataType="DINT" Radix="Decimal" Constant="false" ExternalAccess="Read/Write">
<Data>00 00 00 00</Data>
<Data Format="Decorated">
<DataValue DataType="DINT" Radix="Decimal" Value="0"/>
</Data>
</Tag>""")
        self.set_multilanguage()
        self.assertIsNone(self.tag.description)

    def test_multi_read_none_foreign(self):
        """
        Confirm reading an empty description from a multi-language project
        that has descriptions in other languages.
        """
        self.create_tag("""<Tag Name="tag" TagType="Base" DataType="DINT" Radix="Decimal" Constant="false" ExternalAccess="Read/Write">
<Description>
<LocalizedDescription Lang="zh-CN">
<![CDATA[fail]]>
</LocalizedDescription>
</Description>
<Data>00 00 00 00</Data>
<Data Format="Decorated">
<DataValue DataType="DINT" Radix="Decimal" Value="0"/>
</Data>
</Tag>""")
        self.set_multilanguage()
        self.assertIsNone(self.tag.description)

    def test_single_new(self):
        """Confirm adding a description to a single-language project."""
        self.create_tag("""<Tag Name="tag" TagType="Base" DataType="DINT" Radix="Decimal" Constant="false" ExternalAccess="Read/Write">
<Data>00 00 00 00</Data>
<Data Format="Decorated">
<DataValue DataType="DINT" Radix="Decimal" Value="0"/>
</Data>
</Tag>""")
        self.tag.description = 'new'
        self.assert_description('new')

    def test_multi_new(self):
        """Confirm adding a description to a multi-language project."""
        self.create_tag("""<Tag Name="tag" TagType="Base" DataType="DINT" Radix="Decimal" Constant="false" ExternalAccess="Read/Write">
<Data>00 00 00 00</Data>
<Data Format="Decorated">
<DataValue DataType="DINT" Radix="Decimal" Value="0"/>
</Data>
</Tag>""")
        self.set_multilanguage()
        self.tag.description = 'new'
        self.assert_localized_description('new', self.TARGET_LANGUAGE)

    def test_multi_new_foreign(self):
        """
        Confirm adding a description to a multi-language project that has
        descriptions in other languages.
        """
        self.create_tag("""<Tag Name="tag" TagType="Base" DataType="DINT" Radix="Decimal" Constant="false" ExternalAccess="Read/Write">
<Description>
<LocalizedDescription Lang="zh-CN">
<![CDATA[other]]>
</LocalizedDescription>
</Description>
<Data>00 00 00 00</Data>
<Data Format="Decorated">
<DataValue DataType="DINT" Radix="Decimal" Value="0"/>
</Data>
</Tag>""")
        self.set_multilanguage()
        self.tag.description = 'new'
        self.assert_localized_description('new', self.TARGET_LANGUAGE)
        self.assert_localized_description('other', 'zh-CN')

    def test_single_overwrite(self):
        """
        Confirm overwriting an existing description in a single-language
        project.
        """
        self.create_tag("""<Tag Name="tag" TagType="Base" DataType="DINT" Radix="Decimal" Constant="false" ExternalAccess="Read/Write">
<Description>
<![CDATA[foo]]>
</Description>
<Data>00 00 00 00</Data>
<Data Format="Decorated">
<DataValue DataType="DINT" Radix="Decimal" Value="0"/>
</Data>
</Tag>""")
        self.tag.description = 'new'
        self.assert_description('new')

    def test_multi_overwrite(self):
        """
        Confirm overwriting an existing description in a multi-language
        project.
        """
        self.create_tag("""<Tag Name="tag" TagType="Base" DataType="DINT" Radix="Decimal" Constant="false" ExternalAccess="Read/Write">
<Description>
<LocalizedDescription Lang="en-US">
<![CDATA[old]]>
</LocalizedDescription>
</Description>
<Data>00 00 00 00</Data>
<Data Format="Decorated">
<DataValue DataType="DINT" Radix="Decimal" Value="0"/>
</Data>
</Tag>""")
        self.set_multilanguage()
        self.tag.description = 'new'
        self.assert_localized_description('new', self.TARGET_LANGUAGE)

    def test_multi_overwrite_foreign(self):
        """
        Confirm overwriting an existing description in a multi-language
        project that has descriptions on other languages only affects
        the description in the current language.
        """
        self.create_tag("""<Tag Name="tag" TagType="Base" DataType="DINT" Radix="Decimal" Constant="false" ExternalAccess="Read/Write">
<Description>
<LocalizedDescription Lang="en-US">
<![CDATA[old]]>
</LocalizedDescription>
<LocalizedDescription Lang="zh-CN">
<![CDATA[other]]>
</LocalizedDescription>
</Description>
<Data>00 00 00 00</Data>
<Data Format="Decorated">
<DataValue DataType="DINT" Radix="Decimal" Value="0"/>
</Data>
</Tag>""")
        self.set_multilanguage()
        self.tag.description = 'new'
        self.assert_localized_description('new', self.TARGET_LANGUAGE)
        self.assert_localized_description('other', 'zh-CN')

    def test_single_delete(self):
        """Confirm removing a description from a single-language project."""
        self.create_tag("""<Tag Name="tag" TagType="Base" DataType="DINT" Radix="Decimal" Constant="false" ExternalAccess="Read/Write">
<Description>
<![CDATA[foo]]>
</Description>
<Data>00 00 00 00</Data>
<Data Format="Decorated">
<DataValue DataType="DINT" Radix="Decimal" Value="0"/>
</Data>
</Tag>""")
        self.tag.description = None
        self.assert_no_matching_element('Description')

    def test_multi_delete(self):
        """Confirm removing a description from a multi-language project."""
        self.create_tag("""<Tag Name="tag" TagType="Base" DataType="DINT" Radix="Decimal" Constant="false" ExternalAccess="Read/Write">
<Description>
<LocalizedDescription Lang="en-US">
<![CDATA[old]]>
</LocalizedDescription>
</Description>
<Data>00 00 00 00</Data>
<Data Format="Decorated">
<DataValue DataType="DINT" Radix="Decimal" Value="0"/>
</Data>
</Tag>""")
        self.set_multilanguage()
        self.tag.description = None
        self.assert_no_matching_element('Description')

    def test_multi_delete_foreign(self):
        """
        Confirm removing a description from a multi-language project affects
        only descriptions in the current language.
        """
        self.create_tag("""<Tag Name="tag" TagType="Base" DataType="DINT" Radix="Decimal" Constant="false" ExternalAccess="Read/Write">
<Description>
<LocalizedDescription Lang="en-US">
<![CDATA[old]]>
</LocalizedDescription>
<LocalizedDescription Lang="zh-CN">
<![CDATA[other]]>
</LocalizedDescription>
</Description>
<Data>00 00 00 00</Data>
<Data Format="Decorated">
<DataValue DataType="DINT" Radix="Decimal" Value="0"/>
</Data>
</Tag>""")
        self.set_multilanguage()
        self.tag.description = None

        # Ensure no localized description remains in the current language.
        path = "Description/LocalizedDescription[@Lang='{0}']".format(
            self.TARGET_LANGUAGE)
        self.assert_no_matching_element(path)

        # Ensure descriptions in other languages are unaffected.
        self.assert_localized_description('other', 'zh-CN')

    def assert_description(self, text):
        """
        Verifies a single Description element exists under the Tag and
        contains a matching comment.
        """
        desc = self.tag.element.findall('Description')
        self.assertEqual(len(desc), 1)
        self.assert_cdata_content(desc[0], text)

    def assert_localized_description(self, text, language):
        """
        Verifies a single LocalizedDescription element exists under the
        Description element with a language attribute and matching text.
        """
        path = "Description/LocalizedDescription[@Lang='{0}']".format(language)
        local_desc = self.tag.element.findall(path)
        self.assertEqual(len(local_desc), 1)
        self.assert_cdata_content(local_desc[0], text)


class CommentLanguage(LanguageBase):
    """Tests for multilanguage comments."""
    def test_single_read(self):
        """Confirm reading a comment from a single-language project."""
        self.create_tag("""<Tag Name="tag" TagType="Base" DataType="DINT" Radix="Decimal" Constant="false" ExternalAccess="Read/Write">
<Comments>
<Comment Operand=".0">
<![CDATA[foo]]>
</Comment>
</Comments>
<Data>00 00 00 00</Data>
<Data Format="Decorated">
<DataValue DataType="DINT" Radix="Decimal" Value="0"/>
</Data>
</Tag>""")
        self.assertEqual(self.tag[0].description, 'foo')

    def test_multi_read(self):
        """
        Confirm reading a comment from a multi-language project returns
        only content from the current language.
        """
        self.create_tag("""<Tag Name="tag" TagType="Base" DataType="DINT" Radix="Decimal" Constant="false" ExternalAccess="Read/Write">
<Comments>
<Comment Operand=".0">
<LocalizedComment Lang="en-US">
<![CDATA[pass]]>
</LocalizedComment>
<LocalizedComment Lang="zh-CN">
<![CDATA[fail]]>
</LocalizedComment>
</Comment>
</Comments>
<Data>00 00 00 00</Data>
<Data Format="Decorated">
<DataValue DataType="DINT" Radix="Decimal" Value="0"/>
</Data>
</Tag>""")
        self.set_multilanguage()
        self.assertEqual(self.tag[0].description, 'pass')

    def test_single_read_none(self):
        """
        Confirm reading a nonexistent comment from a single-language project.
        """
        self.create_tag("""<Tag Name="tag" TagType="Base" DataType="DINT" Radix="Decimal" Constant="false" ExternalAccess="Read/Write">
<Data>00 00 00 00</Data>
<Data Format="Decorated">
<DataValue DataType="DINT" Radix="Decimal" Value="0"/>
</Data>
</Tag>""")
        self.assertIsNone(self.tag[0].description)

    def test_multi_read_none_foreign(self):
        """
        Confirm reading a nonexistent comment from a multi-language project
        that has comments in other languages.
        """
        self.create_tag("""<Tag Name="tag" TagType="Base" DataType="DINT" Radix="Decimal" Constant="false" ExternalAccess="Read/Write">
<Comments>
<Comment Operand=".0">
<LocalizedComment Lang="zh-CN">
<![CDATA[other]]>
</LocalizedComment>
</Comment>
</Comments>
<Data>00 00 00 00</Data>
<Data Format="Decorated">
<DataValue DataType="DINT" Radix="Decimal" Value="0"/>
</Data>
</Tag>""")
        self.set_multilanguage()
        self.assertIsNone(self.tag[0].description)

    def test_single_new(self):
        """Confirm adding a comment to a single-language project."""
        self.create_tag("""<Tag Name="tag" TagType="Base" DataType="DINT" Radix="Decimal" Constant="false" ExternalAccess="Read/Write">
<Data>00 00 00 00</Data>
<Data Format="Decorated">
<DataValue DataType="DINT" Radix="Decimal" Value="0"/>
</Data>
</Tag>""")
        self.tag[0].description = 'new'
        self.assert_comment('.0', 'new')

    def test_multi_new(self):
        """Confirm adding a comment to a multi-language project."""
        self.create_tag("""<Tag Name="tag" TagType="Base" DataType="DINT" Radix="Decimal" Constant="false" ExternalAccess="Read/Write">
<Data>00 00 00 00</Data>
<Data Format="Decorated">
<DataValue DataType="DINT" Radix="Decimal" Value="0"/>
</Data>
</Tag>""")
        self.set_multilanguage()
        self.tag[0].description = 'new'
        self.assert_localized_comment('.0', 'new', self.TARGET_LANGUAGE)

    def test_multi_new_foreign(self):
        """
        Confirm adding a comment to a multi-language project that has
        comments in other languages.
        """
        self.create_tag("""<Tag Name="tag" TagType="Base" DataType="DINT" Radix="Decimal" Constant="false" ExternalAccess="Read/Write">
<Comments>
<Comment Operand=".0">
<LocalizedComment Lang="zh-CN">
<![CDATA[other]]>
</LocalizedComment>
</Comment>
</Comments>
<Data>00 00 00 00</Data>
<Data Format="Decorated">
<DataValue DataType="DINT" Radix="Decimal" Value="0"/>
</Data>
</Tag>""")
        self.set_multilanguage()
        self.tag[0].description = 'new'
        self.assert_localized_comment('.0', 'new', self.TARGET_LANGUAGE)
        self.assert_localized_comment('.0', 'other', 'zh-CN')

    def test_single_overwrite(self):
        """
        Confirm overwriting an existing comment in a single-language
        project.
        """
        self.create_tag("""<Tag Name="tag" TagType="Base" DataType="DINT" Radix="Decimal" Constant="false" ExternalAccess="Read/Write">
<Comments>
<Comment Operand=".0">
<![CDATA[foo]]>
</Comment>
</Comments>
<Data>00 00 00 00</Data>
<Data Format="Decorated">
<DataValue DataType="DINT" Radix="Decimal" Value="0"/>
</Data>
</Tag>""")
        self.tag[0].description = 'new'
        self.assert_comment('.0', 'new')

    def test_multi_overwrite(self):
        """
        Confirm overwriting an existing comment in a multi-language
        project.
        """
        self.create_tag("""<Tag Name="tag" TagType="Base" DataType="DINT" Radix="Decimal" Constant="false" ExternalAccess="Read/Write">
<Comments>
<Comment Operand=".0">
<LocalizedComment Lang="en-US">
<![CDATA[old]]>
</LocalizedComment>
</Comment>
</Comments>
<Data>00 00 00 00</Data>
<Data Format="Decorated">
<DataValue DataType="DINT" Radix="Decimal" Value="0"/>
</Data>
</Tag>""")
        self.set_multilanguage()
        self.tag[0].description = 'new'
        self.assert_localized_comment('.0', 'new', self.TARGET_LANGUAGE)

    def test_multi_overwrite_foreign(self):
        """
        Confirm overwriting an existing comment in a multi-language
        project that has comments on other languages only affects
        the comment in the current language.
        """
        self.create_tag("""<Tag Name="tag" TagType="Base" DataType="DINT" Radix="Decimal" Constant="false" ExternalAccess="Read/Write">
<Comments>
<Comment Operand=".0">
<LocalizedComment Lang="en-US">
<![CDATA[old]]>
</LocalizedComment>
<LocalizedComment Lang="zh-CN">
<![CDATA[other]]>
</LocalizedComment>
</Comment>
</Comments>
<Data>00 00 00 00</Data>
<Data Format="Decorated">
<DataValue DataType="DINT" Radix="Decimal" Value="0"/>
</Data>
</Tag>""")
        self.set_multilanguage()
        self.tag[0].description = 'new'
        self.assert_localized_comment('.0', 'new', self.TARGET_LANGUAGE)
        self.assert_localized_comment('.0', 'other', 'zh-CN')

    def test_single_delete(self):
        """Confirm removing a comment from a single-language project."""
        self.create_tag("""<Tag Name="tag" TagType="Base" DataType="DINT" Radix="Decimal" Constant="false" ExternalAccess="Read/Write">
<Comments>
<Comment Operand=".0">
<![CDATA[foo]]>
</Comment>
</Comments>
<Data>00 00 00 00</Data>
<Data Format="Decorated">
<DataValue DataType="DINT" Radix="Decimal" Value="0"/>
</Data>
</Tag>""")
        self.tag[0].description = None
        self.assert_no_matching_element('Comments')

    def test_single_delete_other_operand(self):
        """
        Confirm removing a comment from a single-language project
        does not affect comments for other operands.
        """
        self.create_tag("""<Tag Name="tag" TagType="Base" DataType="DINT" Radix="Decimal" Constant="false" ExternalAccess="Read/Write">
<Comments>
<Comment Operand=".0">
<![CDATA[foo]]>
</Comment>
<Comment Operand=".1">
<![CDATA[bar]]>
</Comment>
</Comments>
<Data>00 00 00 00</Data>
<Data Format="Decorated">
<DataValue DataType="DINT" Radix="Decimal" Value="0"/>
</Data>
</Tag>""")
        self.tag[0].description = None

        # Confirm the target comment was removed.
        self.assert_no_matching_element("Comments/Comment[@Operand='.0']")

        # Confirm the unaffected comment remains.
        self.assert_comment('.1', 'bar')

    def test_single_delete_last(self):
        """
        Confirm removing the last comment in a single-language project
        also removes the overall Comments parent element.
        """
        self.create_tag("""<Tag Name="tag" TagType="Base" DataType="DINT" Radix="Decimal" Constant="false" ExternalAccess="Read/Write">
<Comments>
<Comment Operand=".0">
<![CDATA[foo]]>
</Comment>
<Comment Operand=".1">
<![CDATA[bar]]>
</Comment>
</Comments>
<Data>00 00 00 00</Data>
<Data Format="Decorated">
<DataValue DataType="DINT" Radix="Decimal" Value="0"/>
</Data>
</Tag>""")
        self.tag[0].description = None
        self.tag[1].description = None
        self.assert_no_matching_element('Comments')

    def test_multi_delete(self):
        """Confirm removing a comment from a multi-language project."""
        self.create_tag("""<Tag Name="tag" TagType="Base" DataType="DINT" Radix="Decimal" Constant="false" ExternalAccess="Read/Write">
<Comments>
<Comment Operand=".0">
<LocalizedComment Lang="en-US">
<![CDATA[old]]>
</LocalizedComment>
</Comment>
</Comments>
<Data>00 00 00 00</Data>
<Data Format="Decorated">
<DataValue DataType="DINT" Radix="Decimal" Value="0"/>
</Data>
</Tag>""")
        self.set_multilanguage()
        self.tag[0].description = None
        self.assert_no_matching_element('Comments')

    def test_multi_delete_other_operand(self):
        """
        Confirm removing a comment from a multi-language project
        does not affect comments for other operands.
        """
        self.create_tag("""<Tag Name="tag" TagType="Base" DataType="DINT" Radix="Decimal" Constant="false" ExternalAccess="Read/Write">
<Comments>
<Comment Operand=".0">
<LocalizedComment Lang="en-US">
<![CDATA[foo]]>
</LocalizedComment>
</Comment>
<Comment Operand=".1">
<LocalizedComment Lang="en-US">
<![CDATA[bar]]>
</LocalizedComment>
</Comment>
</Comments>
<Data>00 00 00 00</Data>
<Data Format="Decorated">
<DataValue DataType="DINT" Radix="Decimal" Value="0"/>
</Data>
</Tag>""")
        self.set_multilanguage()
        self.tag[0].description = None

        # Confirm the target comment was removed.
        path = "Comments/Comment[@Operand='.0']" \
               "/LocalizedComment[@Lang='{0}']".format(self.TARGET_LANGUAGE)
        self.assert_no_matching_element(path)

        # Confirm the unaffected comment remains.
        self.assert_localized_comment('.1', 'bar', self.TARGET_LANGUAGE)

    def test_multi_delete_last(self):
        """
        Confirm removing the last comment in a multi-language project
        also removes the overall Comments parent element.
        """
        self.create_tag("""<Tag Name="tag" TagType="Base" DataType="DINT" Radix="Decimal" Constant="false" ExternalAccess="Read/Write">
<Comments>
<Comment Operand=".0">
<LocalizedComment Lang="en-US">
<![CDATA[foo]]>
</LocalizedComment>
</Comment>
<Comment Operand=".1">
<LocalizedComment Lang="en-US">
<![CDATA[bar]]>
</LocalizedComment>
</Comment>
</Comments>
<Data>00 00 00 00</Data>
<Data Format="Decorated">
<DataValue DataType="DINT" Radix="Decimal" Value="0"/>
</Data>
</Tag>""")
        self.set_multilanguage()
        self.tag[0].description = None
        self.tag[1].description = None
        self.assert_no_matching_element('Comments')

    def test_multi_delete_foreign(self):
        """
        Confirm removing a comment from a multi-language project affects
        only comments in the current language.
        """
        self.create_tag("""<Tag Name="tag" TagType="Base" DataType="DINT" Radix="Decimal" Constant="false" ExternalAccess="Read/Write">
<Comments>
<Comment Operand=".0">
<LocalizedComment Lang="en-US">
<![CDATA[foo]]>
</LocalizedComment>
<LocalizedComment Lang="zh-CN">
<![CDATA[other]]>
</LocalizedComment>
</Comment>
</Comments>
<Data>00 00 00 00</Data>
<Data Format="Decorated">
<DataValue DataType="DINT" Radix="Decimal" Value="0"/>
</Data>
</Tag>""")
        self.set_multilanguage()
        self.tag[0].description = None
        self.assert_localized_comment('.0', 'other', 'zh-CN')

    def assert_comment(self, operand, text):
        """
        Verifies a single Comment element exists under the Comments parent
        with a given operand attribute and text content.
        """
        comment = self.get_comment(operand)
        self.assert_cdata_content(comment, text)

    def assert_localized_comment(self, operand, text, language):
        """
        Verifies a single LocalizedComment element exists under the Comments
        element with a language attribute and matching text.
        """
        comment = self.get_comment(operand)
        path = "LocalizedComment[@Lang='{0}']".format(language)
        localized = comment.findall(path)
        self.assertEqual(len(localized), 1)
        self.assert_cdata_content(localized[0], text)

    def get_comment(self, operand):
        """Finds a Comment element with a matching operand attribute."""
        path = "Comments/Comment[@Operand='{0}']".format(operand)
        comment = self.tag.element.findall(path)
        self.assertEqual(len(comment), 1)
        return comment[0]


class AliasFor(unittest.TestCase):
    """Tests for the alias_for attribute of alias tags."""
    def setUp(self):
        e = fixture.parse_xml(r"""<Tag Name="alias" TagType="Alias" Radix="Decimal" AliasFor="tag" ExternalAccess="Read/Write">
<Comments>
<Comment Operand=".0">
<![CDATA[alias operand comment]]>
</Comment>
</Comments>
</Tag>""")
        self.tag = l5x.tag.Tag(e, None)

    def test_alias_read(self):
        """Confirm reading the alias_for attribute returns the correct value."""
        self.assertEqual(self.tag.alias_for,
                         self.tag.element.attrib['AliasFor'])

    def test_alias_write(self):
        """Confirm writing the alias updates the XML attribute."""
        self.tag.alias_for = 'foo'
        self.assertEqual(self.tag.element.attrib['AliasFor'], 'foo')

    def test_remove_comments_on_alias_write(self):
        """Confirm all comments are removed when the alias is changed."""
        self.tag.alias_for = 'bar'
        comments = self.tag.element.find('Comments')
        self.assertIsNone(comments)

    def test_write_non_string_alias(self):
        """Confirm an exception is raised when writing a non-string alias."""
        with self.assertRaises(TypeError):
            self.tag.alias_for = 42

    def test_write_empty_alias(self):
        """Confirm an exception is raised when writing an empty alias string."""
        with self.assertRaises(ValueError):
            self.tag.alias_for = '  '



"""
Unit tests for the top-level project module.
"""

import l5x
import io
import unittest
from tests import fixture
import xml.etree.ElementTree as ElementTree
import xml.dom.minidom


class Parse(unittest.TestCase):
    """Input file parsing tests."""
    def test_invalid_xml(self):
        """Ensure an exception is raised if the XML could not be parsed."""
        buf = io.StringIO(u"foo bar")
        with self.assertRaises(l5x.InvalidFile):
            l5x.Project(buf)

    def test_invalid_root(self):
        """Ensure an exception is raised without the correct root element."""
        doc = ElementTree.Element('random root')

        # Support Python 2 and 3 methods to get a Unicode object.
        try:
            s = ElementTree.tostring(doc, 'unicode')
        except LookupError:
            s = ElementTree.tostring(doc).decode()

        buf = io.StringIO(s)
        with self.assertRaises(l5x.InvalidFile):
            l5x.Project(buf)


class CDATARemoval(unittest.TestCase):
    """Tests for replacing CDATA sections with elements."""
    def setUp(self):
        """
        Creates a dummy project. This is only needed to get a Project
        instance to test the replace_cdata method.
        """
        self.project = fixture.create_project()

    def test_CDATA_to_element(self):
        """Confirm CDATA sections are converted to elements."""
        src = '<root>' + self.generate_cdata('foo') + '</root>'
        root = self.convert_parse(src)
        cdata = root.find('CDATAContent')
        self.assertEqual(cdata.text, 'foo')

    def test_multiple_CDATA(self):
        """Confirm multiple CDATA sections are all converted to elements."""
        src = '<root>' + \
              '<first>' + self.generate_cdata('foo') + '</first>' + \
              '<second>' + self.generate_cdata('bar') + '</second>' + \
              '</root>'
        root = self.convert_parse(src)

        first = root.find('first/CDATAContent')
        self.assertEqual(first.text, 'foo')

        second = root.find('second/CDATAContent')
        self.assertEqual(second.text, 'bar')

    def test_escape_characters(self):
        """Confirm special characters are converted to escape sequences."""
        text = '&<>"\''
        src = '<root>' + self.generate_cdata(text) + '</root>'
        root = self.convert_parse(src)
        cdata = root.find('CDATAContent')
        self.assertEqual(cdata.text, text)

    def test_empty_section(self):
        """Confirm an empty CDATA section is converted to an empty element."""
        src = '<root>' + self.generate_cdata('') + '</root>'
        root = self.convert_parse(src)
        cdata = root.find('CDATAContent')
        self.assertIsNone(cdata.text)

    def test_newline(self):
        """Confirm a CDATA section containing a newline is converted."""
        src = '<root>' + self.generate_cdata('\n') + '</root>'
        root = self.convert_parse(src)
        cdata = root.find('CDATAContent')
        self.assertEqual(cdata.text, '\n')

    def convert_parse(self, src):
        """
        Passes the test string through CDATA removal and parses the
        result into an XML document.
        """
        doc = self.project.convert_to_cdata_element(src)
        return ElementTree.fromstring(doc)

    def generate_cdata(self, text):
        """Creates a CDATA section."""
        return "<![CDATA[{0}]]>".format(text)


class CDATAInsertion(unittest.TestCase):
    """Tests for replacing CDATA elements with CDATA sections."""
    def setUp(self):
        """
        Creates a dummy project. This is only needed to get a Project
        instance to test the CDATA conversion method.
        """
        self.project = fixture.create_project()

    def test_element_to_CDATA(self):
        """Confirm CDATA elements are converted to CDATA sections."""
        src = '<root>' + self.generate_cdata('foo') + '</root>'
        root = self.convert_parse(src)
        self.assert_cdata_content(root, 'foo')

    def test_multiple_elements(self):
        """Confirm multiple CDATA elements are converted to CDATA sections."""
        src = '<root>' + \
              '<first>' + self.generate_cdata('foo') + '</first>' + \
              '<second>' + self.generate_cdata('bar') + '</second>' + \
              '</root>'
        root = self.convert_parse(src)

        for tag, text in [('first', 'foo'), ('second', 'bar')]:
            element = root.getElementsByTagName(tag)[0]
            self.assert_cdata_content(element, text)

    def test_escape_sequences(self):
        """Confirm escape sequences are converted to unescaped characters."""
        src = '<root>' + \
              self.generate_cdata('&amp;&lt;&gt;&quot;&apos;') + \
              '</root>'
        root = self.convert_parse(src)
        self.assert_cdata_content(root, '&<>"\'')

    def test_empty_element(self):
        """Confirm an empty element is converted to an empty CDATA section."""
        src = '<root>' + self.generate_cdata('') + '</root>'
        self.assert_empty_cdata(src)

    def test_newline(self):
        """Confirm an element containing a newline is converted."""
        src = '<root>' + self.generate_cdata('\n') + '</root>'
        root = self.convert_parse(src)
        self.assert_cdata_content(root, '\n')

    def test_self_closing_empty(self):
        """Confirm a self-closing element yields to an empty CDATA section."""
        src = '<root><CDATAContent/></root>'
        self.assert_empty_cdata(src)

    def test_tag_whitespace(self):
        """Confirm whitespace in opening and closing tags is accepted."""
        src = '<root><CDATAContent   ></CDATAContent   ></root>'
        self.assert_empty_cdata(src)

    def test_self_closing_tag_whitespace(self):
        """Confirm whitespace in a self-closing tag is accepted."""
        src = '<root><CDATAContent   /></root>'
        self.assert_empty_cdata(src)

    def convert_parse(self, src):
        """
        Converts string containing the source document, parses the
        result, and returns the root element.
        """
        cdata = self.project.convert_to_cdata_section(src)
        doc = xml.dom.minidom.parseString(cdata)
        return doc.documentElement

    def generate_cdata(src, text):
        """Creates a CDATAContent element containing text."""
        return "<CDATAContent>{0}</CDATAContent>".format(text)
        
    def assert_cdata_content(self, element, text):
        """
        Confirms an element contains a single CDATA section with a
        given text content.
        """
        cdata_nodes = [n for n in element.childNodes
                       if n.nodeType == n.CDATA_SECTION_NODE]
        self.assertEqual(len(cdata_nodes), 1)
        self.assertEqual(cdata_nodes[0].data, text)

    def assert_empty_cdata(self, src):
        """Validates CDATA conversion results in an empty CDATA section.

        This uses basic string equality instead of parsing the converted
        XML because empty CDATA sections are not preserved when parsed.
        """
        doc = self.project.convert_to_cdata_section(src)
        self.assertEqual(doc, '<root><![CDATA[]]></root>')


class ContentElements(unittest.TestCase):
    """Tests to verify access to top-level content elements."""
    def test_controller(self):
        """Confirm access to the controller."""
        prj = fixture.create_project(self.add_mock_controller_path)
        self.assertEqual(prj.controller.comm_path, 'this is the controller')

    def add_mock_controller_path(self, doc):
        """Creates a dummy controller path for the controller test case."""
        controller = doc.getElementsByTagName('Controller')[0]
        controller.setAttribute('CommPath', 'this is the controller')

    def test_programs(self):
        """Confirm access to the set of programs."""
        prj = fixture.create_project(self.add_mock_program)
        prj.programs['Some Program']

    def add_mock_program(self, doc):
        """Creates a dummy program for the programs test case."""
        parent = doc.getElementsByTagName('Programs')[0]
        prog = doc.createElement('Program')
        prog.setAttribute('Name', 'Some Program')
        parent.appendChild(prog)

    def test_modules(self):
        """Confirm access to the set of I/O modules."""
        prj = fixture.create_project(self.add_mock_module)
        prj.modules['SpamModule']

    def add_mock_module(self, doc):
        """Creates a dummy module for the modules test case."""
        parent = doc.getElementsByTagName('Modules')[0]
        mod = doc.createElement('Module')
        mod.setAttribute('Name', 'SpamModule')
        parent.appendChild(mod)



from setuptools import setup

setup(name='l5x',
      version='1.5',
      author='Jason Valenzuela',
      author_email='jvalenzuela1977@gmail.com',
      packages=['l5x', 'tests'],
      url='https://github.com/jvalenzuela/l5x',
      license='LICENSE.txt',
      description='RSLogix .L5X interface.',
      long_description=open('README.rst').read(),
      classifiers=[
        'Development Status :: 5 - Production/Stable',
        'License :: OSI Approved :: GNU General Public License v3 or later (GPLv3+)',
        'Intended Audience :: Developers',
        'Operating System :: OS Independent',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3',
        'Topic :: Text Processing :: Markup :: XML',
        'Topic :: Software Development :: Libraries :: Python Modules',
        'Topic :: Scientific/Engineering :: Interface Engine/Protocol Translator'
        ])



"""
Unit tests for low-level XML interface objects.
"""

from l5x import dom
import unittest
import xml.etree.ElementTree as ElementTree


class ElementDict(unittest.TestCase):
    """Unit tests for the ElementDict class."""
    class Dummy(object):
        """Mock dictionary value type."""
        def __init__(self, element, *args):
            self.element = element
            self.args = args

    def test_key_attribute_value(self):
        """Confirm values from the key attribute are used for lookup."""
        parent = ElementTree.Element('parent')
        child = ElementTree.SubElement(parent, 'child', {'key':'foo'})
        d = dom.ElementDict(parent, 'key', self.Dummy)
        self.assertIs(d['foo'].element, child)

    def test_depth_lookup(self):
        """Confirm only direct children are queried for lookup."""
        parent = ElementTree.Element('parent')
        child = ElementTree.SubElement(parent, 'child', {'key':'foo'})
        ElementTree.SubElement(child, 'grandchild', {'key':'bar'})
        d = dom.ElementDict(parent, 'key', self.Dummy)
        with self.assertRaises(KeyError):
            d['bar']

    def test_key_not_found(self):
        """Confirm a KeyError is raised if a matching key doesn't exist."""
        parent = ElementTree.Element('parent')
        ElementTree.SubElement(parent, 'child', {'key':'foo'})
        d = dom.ElementDict(parent, 'key', self.Dummy)
        with self.assertRaises(KeyError):
            d['bar']

    def test_key_not_found_empty(self):
        """Confirm a KeyError is raised if the parent has no child elements."""
        parent = ElementTree.Element('parent')
        d = dom.ElementDict(parent, 'key', self.Dummy)
        with self.assertRaises(KeyError):
            d['bar']

    def test_string_lookup(self):
        """Confirm correct lookup with string keys."""
        parent = ElementTree.Element('parent')
        child = ElementTree.SubElement(parent, 'child', {'key':'foo'})
        d = dom.ElementDict(parent, 'key', self.Dummy)
        self.assertIs(d['foo'].element, child)

    def test_integer_lookup(self):
        """Confirm correct lookup with integer keys."""
        parent = ElementTree.Element('parent')
        child = ElementTree.SubElement(parent, 'child', {'key':'42'})
        d = dom.ElementDict(parent, 'key', self.Dummy, key_type=int)
        self.assertIs(d[42].element, child)

    def test_value_read_only(self):
        """
        Confirm attempting to assign a different value raises an exception.
        """
        parent = ElementTree.Element('parent')
        ElementTree.SubElement(parent, 'child', {'key':'foo'})
        d = dom.ElementDict(parent, 'key', self.Dummy)
        with self.assertRaises(TypeError):
            d['foo'] = 0

    def test_create_new(self):
        """
        Confirm attempting to create a new key raises an exception.
        """
        parent = ElementTree.Element('parent')
        d = dom.ElementDict(parent, 'key', self.Dummy)
        with self.assertRaises(TypeError):
            d['bar'] = 'foo'

    def test_names(self):
        """Confirm the names attribute returns a list of keys."""
        parent = ElementTree.Element('parent')
        ElementTree.SubElement(parent, 'child', {'key':'foo'})
        ElementTree.SubElement(parent, 'child', {'key':'bar'})
        ElementTree.SubElement(parent, 'child', {'key':'baz'})
        d = dom.ElementDict(parent, 'key', self.Dummy)
        names = set(d.names)
        self.assertEqual(names, set(['foo', 'bar', 'baz']))

    def test_names_empty(self):
        """
        Confirm the names attribute returns an empty list when the parent has
        no child elements.
        """
        parent = ElementTree.Element('parent')
        d = dom.ElementDict(parent, 'key', self.Dummy)
        self.assertFalse(d.names)

    def test_names_attribute(self):
        """Confirm the key attribute is used to generate the names list."""
        parent = ElementTree.Element('parent')
        ElementTree.SubElement(parent, 'child', {'key':'foo', 'attr':'spam'})
        ElementTree.SubElement(parent, 'child', {'key':'bar', 'attr':'eggs'})
        d = dom.ElementDict(parent, 'key', self.Dummy)
        names = set(d.names)
        self.assertEqual(names, set(['foo', 'bar']))

    def test_names_type(self):
        """Confirm names are converted to the correct key type."""
        parent = ElementTree.Element('parent')
        child = ElementTree.SubElement(parent, 'child', {'key':'42'})
        d = dom.ElementDict(parent, 'key', self.Dummy, key_type=int)
        self.assertIsInstance(d.names[0], int)

    def test_names_read_only(self):
        """
        Confirm attempting to assign the names attributes raises an exception.
        """
        parent = ElementTree.Element('parent')
        ElementTree.SubElement(parent, 'child', {'key':'foo'})
        d = dom.ElementDict(parent, 'key', self.Dummy)
        with self.assertRaises(AttributeError):
            d.names = 'foo'

    def test_single_value_type(self):
        """
        Confirm an instance of the value type is returned when the value
        type is specified as a class.
        """
        parent = ElementTree.Element('parent')
        ElementTree.SubElement(parent, 'child', {'key':'foo'})
        d = dom.ElementDict(parent, 'key', self.Dummy)
        self.assertIsInstance(d['foo'], self.Dummy)

    def test_single_value_type_args(self):
        """
        Confirm the target element and extra arguments are passed to
        the value class for single-type values.
        """
        parent = ElementTree.Element('parent')
        child = ElementTree.SubElement(parent, 'child', {'key':'foo'})
        d = dom.ElementDict(parent, 'key', self.Dummy,
                            value_args=['spam', 'eggs'])
        self.assertIs(d['foo'].element, child)
        self.assertEqual(d['foo'].args, ('spam', 'eggs'))

    def test_value_type_by_attribute(self):
        """
        Confirm the type of value returned is selected by the type attribute
        when the value type is specified as a dictionary.
        """
        class Foo(self.Dummy): pass
        class Bar(self.Dummy): pass
        types = {'Foo':Foo, 'Bar':Bar}
        parent = ElementTree.Element('parent')
        ElementTree.SubElement(parent, 'child', {'key':'foo', 'type':'Foo'})
        ElementTree.SubElement(parent, 'child', {'key':'bar', 'type':'Bar'})
        d = dom.ElementDict(parent, 'key', types, type_attr='type')
        self.assertIsInstance(d['foo'], Foo)
        self.assertIsInstance(d['bar'], Bar)

    def test_value_type_by_attribute_args(self):
        """
        Confirm the target element and extra arguments are passed to
        the value class for value types selected by attribute.
        """
        class Foo(self.Dummy): pass
        types = {'Foo':Foo}
        parent = ElementTree.Element('parent')
        child = ElementTree.SubElement(parent, 'child',
                                       {'key':'foo', 'type':'Foo'})
        d = dom.ElementDict(parent, 'key', types, type_attr='type',
                        value_args=['spam', 'eggs'])
        self.assertIs(d['foo'].element, child)
        self.assertEqual(d['foo'].args, ('spam', 'eggs'))


class CDATAElement(unittest.TestCase):
    """Unit tests for the CDATAElement class."""
    def test_existing_get_string(self):
        """Confirm CDATA element content is retrieved as the object's string."""
        e = ElementTree.Element('element')
        cdata = ElementTree.SubElement(e, dom.CDATA_TAG)
        cdata.text = 'foo'
        self.assertEqual(str(dom.CDATAElement(e)), 'foo')

    def test_existing_set_string(self):
        """Confirm CDATA element content is updated when set."""
        e = ElementTree.Element('element')
        ElementTree.SubElement(e, dom.CDATA_TAG)
        cdata = dom.CDATAElement(e)
        cdata.set('foo')
        self.assertEqual(str(cdata), 'foo')

    def test_existing_empty(self):
        """Confirm an existing empty element returns an empty string."""
        e = ElementTree.Element('element')
        child = ElementTree.SubElement(e, dom.CDATA_TAG)
        cdata = dom.CDATAElement(e)
        self.assertEqual(str(cdata), '')

    def test_new_cdata_child(self):
        """Confirm a single CDATA child is included when creating an element."""
        parent = ElementTree.Element('parent')
        dom.CDATAElement(parent=parent, name='new')
        cdata = parent.find('new')
        children = cdata.findall('*')
        self.assertEqual(len(children), 1)
        self.assertEqual(children[0].tag, dom.CDATA_TAG)

    def test_new_tag(self):
        """Confirm tag name when creating a new element."""
        parent = ElementTree.Element('parent')
        dom.CDATAElement(parent=parent, name='new')
        children = parent.findall('*')
        self.assertEqual(children[0].tag, 'new')

    def test_new_parent(self):
        """Confirm placement under parent when creating a new element."""
        parent = ElementTree.Element('parent')
        dom.CDATAElement(parent=parent, name='new')
        children = parent.findall('*')
        self.assertNotEqual(len(children), 0)

    def test_new_attributes(self):
        """Confirm attribute assignment when creating a new element."""
        attrs = {'foo':'bar', 'spam':'eggs'}
        parent = ElementTree.Element('parent')
        dom.CDATAElement(parent=parent, name='new', attributes=attrs)
        child = parent.find('new')
        self.assertEqual(child.attrib, attrs)


class AttributeDescriptor(unittest.TestCase):
    """Unit tests for the AttributeDescriptor class."""
    class Dummy(object):
        """Test fixture."""
        read_write = dom.AttributeDescriptor('read_write')
        read_only = dom.AttributeDescriptor('read_only', True)
        nonexistent = dom.AttributeDescriptor('nonexistent')

        def __init__(self, element):
            self.element = element

    def setUp(self):
        attrs = {'read_write':'foo',
                 'read_only':'bar'}
        self.element = ElementTree.Element('tag', attrs)
        self.dummy = self.Dummy(self.element)

    def test_get(self):
        """Confirm a get request returns the target attribute's value."""
        self.assertEqual(self.dummy.read_write, 'foo')

    def test_get_nonexistent(self):
        """Confirm getting a nonexistent attribute returns None."""
        self.assertIsNone(self.dummy.nonexistent)

    def test_set(self):
        """Confirm assignment updates the target attribute's value."""
        self.dummy.read_write = 'spam'
        self.assertEqual(self.element.attrib['read_write'], 'spam')

    def test_new(self):
        """
        Confirm a new attribute is created when assigning a value to a
        nonexistent attribute.
        """
        self.dummy.nonexistent = 'foo'
        self.assertEqual(self.element.attrib['nonexistent'], 'foo')

    def test_set_read_only(self):
        """Confirm assignment to a read-only attribute raises an exception."""
        with self.assertRaises(AttributeError):
            self.dummy.read_only = 'spam'

    def test_delete(self):
        """Confirm attribute is removed when set to None."""
        self.dummy.read_write = None
        self.assertFalse('read_write' in self.element.attrib.keys())

    def test_delete_nonexistent(self):
        """Confirm deleting an attribute that does not exist does not fail."""
        self.dummy.nonexistent = None

    def test_delete_read_only(self):
        """Confirm deleting a read-only attribute raises an exception."""
        with self.assertRaises(AttributeError):
            self.dummy.read_only = None

    def test_invalid_value_type(self):
        """Confirm assigning a non-string value raises an exception."""
        with self.assertRaises(TypeError):
            self.dummy.read_write = 42
